from typing import Annotated, Optional

import typer

from jobflow_remote.cli.jf import app
from jobflow_remote.cli.jfr_typer import JFRTyper
from jobflow_remote.jobs.run import run_batch_jobs, run_remote_job

app_execution = JFRTyper(
    name="execution",
    help="Commands for executing the jobs locally",
    no_args_is_help=True,
    hidden=True,
)
app.add_typer(app_execution)


@app_execution.command()
def run(
    run_dir: Annotated[
        Optional[str],
        typer.Argument(
            help="The path to the folder where the files of the job to run will be executed",
        ),
    ] = ".",
) -> None:
    """Run the Job in the selected folder based on the."""
    run_remote_job(run_dir)


@app_execution.command()
def run_batch(
    base_run_dir: Annotated[
        str,
        typer.Argument(
            help="The path to the base folder where the jobs will be executed",
        ),
    ],
    files_dir: Annotated[
        str,
        typer.Argument(
            help="The path to the folder where files for handling the batch jobs will be stored",
        ),
    ],
    process_uuid: Annotated[
        str,
        typer.Argument(
            help="A uuid representing the batch process",
        ),
    ],
    max_time: Annotated[
        Optional[int],
        typer.Option(
            "--max-time",
            "-mt",
            help=(
                "The maximum time after which no more jobs will be executed (seconds)"
            ),
        ),
    ] = None,
    max_wait: Annotated[
        Optional[int],
        typer.Option(
            "--max-wait",
            "-mw",
            help=(
                "The maximum time the job will wait before stopping if no jobs are available to run (seconds)"
            ),
        ),
    ] = 60,
    max_jobs: Annotated[
        Optional[int],
        typer.Option(
            "--max-jobs",
            "-mj",
            help=("The maximum number of jobs that will be executed by the batch job"),
        ),
    ] = None,
    parallel_jobs: Annotated[
        Optional[int],
        typer.Option(
            "--parallel-jobs",
            "-pj",
            help=("Number of jobs executed in parallel"),
        ),
    ] = None,
) -> None:
    """Run Jobs in batch mode."""
    run_batch_jobs(
        base_run_dir,
        files_dir,
        process_uuid,
        max_time=max_time,
        max_wait=max_wait,
        max_jobs=max_jobs,
        parallel_jobs=parallel_jobs,
    )



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from qtoolkit.core.data_objects import CancelResult, QJob, QResources, SubmissionResult
from qtoolkit.io.shell import ShellIO

if TYPE_CHECKING:
    from collections.abc import Sequence

    from qtoolkit.io.base import BaseSchedulerIO

    from jobflow_remote.remote.host import BaseHost

OUT_FNAME = "queue.out"
ERR_FNAME = "queue.err"


def set_name_out(
    resources: dict | QResources,
    name: str,
    out_fpath: str | Path = OUT_FNAME,
    err_fpath: str | Path = ERR_FNAME,
) -> None:
    # sanitize the name
    name = name.replace(" ", "_")
    if isinstance(resources, QResources):
        resources.job_name = name
        resources.output_filepath = out_fpath
        resources.error_filepath = err_fpath
    else:
        resources["job_name"] = name
        resources["qout_path"] = out_fpath
        resources["qerr_path"] = err_fpath


class QueueManager:
    """Base class for job queues.

    Attributes
    ----------
    scheduler_io : str
        Name of the queue
    host : BaseHost
        Host where the command should be executed.
    """

    def __init__(
        self,
        scheduler_io: BaseSchedulerIO,
        host: BaseHost,
        timeout_exec: int | None = None,
    ) -> None:
        self.scheduler_io = scheduler_io
        self.host = host
        self.timeout_exec = timeout_exec

    def execute_cmd(
        self, cmd: str, workdir: str | Path | None = None, timeout: int | None = None
    ):
        """Execute a command.

        Parameters
        ----------
        cmd : str
            Command to be executed
        workdir: str or None
            path where the command will be executed.

        Returns
        -------
        stdout : str
        stderr : str
        exit_code : int
        """
        timeout = timeout if timeout is not None else self.timeout_exec
        return self.host.execute(cmd, workdir, timeout)

    def get_submission_script(
        self,
        commands: str | list[str] | None,
        options: dict | QResources | None = None,
        work_dir: str | Path | None = None,
        pre_run: str | list[str] | None = None,
        post_run: str | list[str] | None = None,
        export: dict | None = None,
        modules: list[str] | None = None,
    ) -> str:
        """ """
        commands_list = []
        if change_dir := self.get_change_dir(work_dir):
            commands_list.append(change_dir)
        if pre_run := self.get_pre_run(pre_run):
            commands_list.append(pre_run)
        if export_str := self.get_export(export):
            commands_list.append(export_str)
        if modules_str := self.get_modules(modules):
            commands_list.append(modules_str)
        if run_commands := self.get_run_commands(commands):
            commands_list.append(run_commands)
        if post_run := self.get_post_run(post_run):
            commands_list.append(post_run)
        return self.scheduler_io.get_submission_script(commands_list, options)

    def get_change_dir(self, dir_path: str | Path | None) -> str:
        if dir_path:
            return f"cd {dir_path}"
        return ""

    def get_pre_run(self, pre_run: str | list[str] | None) -> str:
        if isinstance(pre_run, (list, tuple)):
            return "\n".join(pre_run)
        return pre_run

    def get_export(self, exports: dict | None) -> str | None:
        if not exports:
            return None
        exports_str = []
        for k, v in exports.items():
            exports_str.append(f"export {k}={v}")
        return "\n".join(exports_str)

    def get_modules(self, modules: list[str] | None) -> str | None:
        if not modules:
            return None
        modules_str = [f"module load {m}" for m in modules]
        return "\n".join(modules_str)

    def get_run_commands(self, commands) -> str:
        if isinstance(commands, str):
            return commands
        if isinstance(commands, list):
            return "\n".join(commands)
        raise ValueError("commands should be a str or a list of str.")

    def get_post_run(self, post_run: str | list[str] | None) -> str:
        if isinstance(post_run, (list, tuple)):
            return "\n".join(post_run)
        return post_run

    def submit(
        self,
        commands: str | list[str] | None,
        options=None,
        work_dir=None,
        pre_run: str | list[str] | None = None,
        post_run: str | list[str] | None = None,
        export: dict | None = None,
        modules: list[str] | None = None,
        script_fname="submit.sh",
        create_submit_dir: bool = False,
        timeout: int | None = None,
    ) -> SubmissionResult:
        script_fpath = self.write_submission_script(
            commands=commands,
            options=options,
            work_dir=work_dir,
            pre_run=pre_run,
            post_run=post_run,
            export=export,
            modules=modules,
            script_fname=script_fname,
            create_submit_dir=create_submit_dir,
        )
        submit_cmd = self.scheduler_io.get_submit_cmd(script_fpath)
        stdout, stderr, returncode = self.execute_cmd(
            submit_cmd, work_dir, timeout=timeout
        )
        return self.scheduler_io.parse_submit_output(
            exit_code=returncode, stdout=stdout, stderr=stderr
        )

    def write_submission_script(
        self,
        commands: str | list[str] | None,
        options=None,
        work_dir=None,
        pre_run: str | list[str] | None = None,
        post_run: str | list[str] | None = None,
        export: dict | None = None,
        modules: list[str] | None = None,
        script_fname="submit.sh",
        create_submit_dir: bool = False,
    ):
        script_str = self.get_submission_script(
            commands=commands,
            options=options,
            work_dir=work_dir,
            pre_run=pre_run,
            post_run=post_run,
            export=export,
            modules=modules,
        )
        if create_submit_dir and work_dir:
            created = self.host.mkdir(work_dir, recursive=True, exist_ok=True)
            if not created:
                raise RuntimeError("failed to create directory")
        script_fpath = Path(work_dir, script_fname) if work_dir else Path(script_fname)
        self.host.write_text_file(script_fpath, script_str)
        return script_fpath

    def cancel(self, job: QJob | int | str, timeout: int | None = None) -> CancelResult:
        cancel_cmd = self.scheduler_io.get_cancel_cmd(job)
        stdout, stderr, returncode = self.execute_cmd(cancel_cmd, timeout=timeout)
        return self.scheduler_io.parse_cancel_output(
            exit_code=returncode, stdout=stdout, stderr=stderr
        )

    def get_job(self, job: QJob | int | str, timeout: int | None = None) -> QJob | None:
        job_cmd = self.scheduler_io.get_job_cmd(job)
        stdout, stderr, returncode = self.execute_cmd(job_cmd, timeout=timeout)
        return self.scheduler_io.parse_job_output(
            exit_code=returncode, stdout=stdout, stderr=stderr
        )

    def get_jobs_list(
        self,
        jobs: Sequence[QJob | int | str] | None = None,
        user: str | None = None,
        timeout: int | None = None,
    ) -> list[QJob]:
        # in order to avoid issues with schedulers that do not support query by
        # list of job ids, if the user is passed ignore the job ids.
        if user is not None:
            jobs = None
        job_cmd = self.scheduler_io.get_jobs_list_cmd(jobs, user)
        stdout, stderr, returncode = self.execute_cmd(job_cmd, timeout=timeout)
        return self.scheduler_io.parse_jobs_list_output(
            exit_code=returncode, stdout=stdout, stderr=stderr
        )

    def get_shell_manager(self):
        return QueueManager(
            scheduler_io=ShellIO(), host=self.host, timeout_exec=self.timeout_exec
        )



from __future__ import annotations

import importlib.metadata
import json
import logging
import traceback
from typing import TYPE_CHECKING

from jobflow_remote.config.base import (
    ExecutionConfig,
    LocalWorker,
    Project,
    RemoteWorker,
    WorkerBase,
)

if TYPE_CHECKING:
    from jobflow import JobStore
    from maggma.core import Store

    from jobflow_remote.remote.host import BaseHost

logger = logging.getLogger(__name__)


def generate_dummy_project(name: str, full: bool = False) -> Project:
    remote_worker = generate_dummy_worker(scheduler_type="slurm", host_type="remote")
    workers = {"example_worker": remote_worker}
    exec_config = {}
    if full:
        local_worker = generate_dummy_worker(scheduler_type="shell", host_type="local")
        workers["example_local"] = local_worker
        exec_config = {"example_config": generate_dummy_exec_config()}

    queue = {"store": generate_dummy_queue()}

    jobstore = generate_dummy_jobstore()

    return Project(
        name=name,
        jobstore=jobstore,
        queue=queue,
        workers=workers,
        exec_config=exec_config,
    )


def generate_dummy_worker(
    scheduler_type: str = "slurm", host_type: str = "remote"
) -> WorkerBase:
    d: dict = dict(
        scheduler_type=scheduler_type,
        work_dir="/path/to/run/folder",
        pre_run="source /path/to/python/environment/activate",
    )
    if host_type == "local":
        d.update(
            type="local",
            timeout_execute=60,
        )
        return LocalWorker(**d)
    if host_type == "remote":
        d.update(
            type="remote",
            host="remote.host.net",
            user="bob",
            timeout_execute=60,
        )
        return RemoteWorker(**d)

    raise ValueError(f"Unknown/unhandled host type: {host_type}")


def generate_dummy_jobstore() -> dict:
    return {
        "docs_store": {
            "type": "MongoStore",
            "database": "db_name",
            "host": "host.mongodb.com",
            "port": 27017,
            "username": "bob",
            "password": "secret_password",
            "collection_name": "outputs",
        },
        "additional_stores": {
            "data": {
                "type": "GridFSStore",
                "database": "db_name",
                "host": "host.mongodb.com",
                "port": 27017,
                "username": "bob",
                "password": "secret_password",
                "collection_name": "outputs_blobs",
            }
        },
    }


def generate_dummy_exec_config() -> ExecutionConfig:
    return ExecutionConfig(
        modules=["GCC/10.2.0", "OpenMPI/4.0.5-GCC-10.2.0"],
        export={"PATH": "/path/to/binaries:$PATH"},
        pre_run="conda activate env_name",
    )


def generate_dummy_queue() -> dict:
    return dict(
        type="MongoStore",
        host="localhost",
        database="db_name",
        username="bob",
        password="secret_password",
        collection_name="jobs",
    )


def _check_workdir(worker: WorkerBase, host: BaseHost) -> str | None:
    """Check that the configured workdir exists or is writable on the worker.

    Parameters
    ----------
        worker: The worker configuration.
        host: A connected host.

    """
    try:
        host_error = host.test()
        if host_error:
            return host_error
    except Exception:
        exc = traceback.format_exc()
        return f"Error while testing worker:\n {exc}"

    canary_file = worker.work_dir / ".jf_heartbeat"
    try:
        # First try to create the folder. The runner will create is anyway and
        # it should be less confusing for the user.
        host.mkdir(worker.work_dir)
        host.write_text_file(canary_file, "\n")
        return None  # noqa: TRY300
    except FileNotFoundError as exc:
        raise FileNotFoundError(
            f"Could not write to {canary_file}. Does the folder exist on the remote?\nThe folder should be specified as an absolute path with no shell expansions or environment variables."
        ) from exc
    except PermissionError as exc:
        raise PermissionError(
            f"Could not write to {canary_file}. Do you have the rights to access that folder?"
        ) from exc
    finally:
        # Must be enclosed in quotes with '!r' as the path may contain spaces
        host.execute(f"rm {str(canary_file)!r}")


def check_worker(
    worker: WorkerBase, full_check: bool = False
) -> tuple[str | None, str | None]:
    """Check that a connection to the configured worker can be made."""
    host = worker.get_host()
    worker_warn = None
    try:
        host.connect()
        host_error = host.test()
        if host_error:
            return host_error, None

        from jobflow_remote.remote.queue import QueueManager

        qm = QueueManager(scheduler_io=worker.get_scheduler_io(), host=host)
        qm.get_jobs_list()

        workdir_err = _check_workdir(worker=worker, host=host)
        if workdir_err:
            return workdir_err, None

        # don't perform the environment check, as they will be equivalent
        if worker.type != "local":
            worker_warn = _check_environment(
                worker=worker, host=host, full_check=full_check
            )

    except Exception:
        exc = traceback.format_exc()
        return f"Error while testing worker:\n {exc}", worker_warn
    finally:
        try:
            host.close()
        except Exception:
            logger.warning(f"error while closing connection to host {host}")

    return None, worker_warn


def _check_store(store: Store) -> str | None:
    try:
        store.connect()
        store.query_one()
    except Exception:
        return traceback.format_exc()
    finally:
        store.close()

    return None


def check_queue_store(queue_store: Store) -> str | None:
    err = _check_store(queue_store)
    if err:
        return f"Error while checking queue store:\n{err}"
    return None


def check_jobstore(jobstore: JobStore) -> str | None:
    err = _check_store(jobstore.docs_store)
    if err:
        return f"Error while checking docs_store store:\n{err}"
    for store_name, store in jobstore.additional_stores.items():
        err = _check_store(store)
        if err:
            return f"Error while checking additional store {store_name}:\n{err}"
    return None


def _check_environment(
    worker: WorkerBase, host: BaseHost, full_check: bool = False
) -> str | None:
    """Check that the worker has a python environment with the same versions of libraries.

    Parameters
    ----------
        host: A connected host.
        full_check: Whether to check the entire environment and not just jobflow and jobflow-remote.

    Returns
    -------
    str | None
        A message describing the environment mismatches. None if no mismatch is found.
    """
    installed_packages = importlib.metadata.distributions()
    local_package_versions = {
        package.metadata["Name"]: package.version for package in installed_packages
    }
    cmd = "pip list --format=json"
    if worker.pre_run:
        cmd = "; ".join(worker.pre_run.strip().splitlines()) + "; " + cmd

    stdout, stderr, errcode = host.execute(cmd)
    if errcode != 0:
        return f"Error while checking the compatibility of the environments: {stderr}"
    host_package_versions = {
        package_dict["name"]: package_dict["version"]
        for package_dict in json.loads(stdout)
    }
    if full_check:
        packages_to_check = list(local_package_versions.keys())
    else:
        packages_to_check = ["jobflow", "jobflow-remote"]
    missing = []
    mismatch = []
    for package in packages_to_check:
        if package not in host_package_versions:
            missing.append((package, local_package_versions[package]))
            continue
        if local_package_versions[package] != host_package_versions[package]:
            mismatch.append(
                (
                    package,
                    local_package_versions[package],
                    host_package_versions[package],
                )
            )
    msg = None
    if mismatch or missing:
        msg = "Note: inconsistencies may be due to the proper python environment not being correctly loaded.\n"
    if missing:
        missing_str = [f"{m[0]} - {m[1]}" for m in missing]
        msg += f"Missing packages: {', '.join(missing_str)}. "
    if mismatch:
        mismatch_str = [f"{m[0]} - {m[1]} vs {m[2]}" for m in mismatch]
        msg += f"Mismatching versions: {', '.join(mismatch_str)}"

    return msg



"""Tools for logging."""

from __future__ import annotations

import logging
import logging.config
from pathlib import Path

from monty.os import makedirs_p


def initialize_runner_logger(
    log_folder: str | Path, level: int = logging.INFO, runner_id: str | None = None
) -> None:
    """
    Initialize the runner logger.

    Parameters
    ----------
    level
        The log level.
    """
    # If the directory it is not present it does not initialize the logger with
    # an unclear error message. Since this is only for the runner logger
    # it should not be a problem to check that the directory exists when the
    # runner is started.
    makedirs_p(log_folder)

    if runner_id:
        msg_format = f"%(asctime)s [%(levelname)s] ID {runner_id} %(name)s: %(message)s"
    else:
        msg_format = "%(asctime)s [%(levelname)s] %(name)s: %(message)s"

    config = {
        "version": 1,
        "disable_existing_loggers": True,
        "formatters": {
            "standard": {"format": msg_format},
        },
        "handlers": {
            "default": {
                "class": "logging.handlers.RotatingFileHandler",
                "level": level,
                "formatter": "standard",
                "filename": str(Path(log_folder) / "runner.log"),
                "mode": "a",
                "backupCount": 5,
                "encoding": "utf8",
                "maxBytes": 5000000,
            },
            "stream": {
                "level": level,
                "formatter": "standard",
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",  # Default is stderr
            },
        },
        "loggers": {
            "jobflow_remote": {  # root logger
                # 'handlers': ['default'],
                "handlers": ["default", "stream"],
                "level": level,
                "propagate": False,
            },
        },
    }

    logging.config.dictConfig(config)


def initialize_cli_logger(
    level: int = logging.WARNING, full_exc_info: bool = True
) -> None:
    """
    Initialize the logger for the CLI based on rich.

    Parameters
    ----------
    level
        The log level.
    """
    config = {
        "version": 1,
        "disable_existing_loggers": True,
        "formatters": {
            "cli_formatter": {
                "()": lambda: CLIFormatter(
                    log_exception_trace=full_exc_info, datefmt="[%X]"
                ),
            }
        },
        "handlers": {
            "rich": {
                "class": "rich.logging.RichHandler",
                "level": level,
                "formatter": "cli_formatter",
                "show_path": False,
            },
        },
        "loggers": {
            "jobflow_remote": {  # root logger
                "handlers": ["rich"],
                "level": level,
                "propagate": False,
            },
        },
    }

    logging.config.dictConfig(config)


def initialize_remote_run_log(level: int = logging.INFO) -> None:
    """
    Initialize the logger for the execution of the jobs.

    Parameters
    ----------
    level
        The log level.
    """
    config = {
        "version": 1,
        "disable_existing_loggers": True,
        "formatters": {
            "standard": {"format": "%(asctime)s [%(levelname)s] %(name)s: %(message)s"},
        },
        "handlers": {
            "stream": {
                "level": level,
                "formatter": "standard",
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",  # Default is stderr
            },
        },
        "loggers": {
            "jobflow_remote": {  # root logger
                "handlers": ["stream"],
                "level": level,
                "propagate": False,
            },
        },
    }

    logging.config.dictConfig(config)


class CLIFormatter(logging.Formatter):
    def __init__(self, log_exception_trace: bool = True, **kwargs) -> None:
        super().__init__(**kwargs)
        self.log_exception_trace = log_exception_trace

    def formatException(self, ei):
        if self.log_exception_trace:
            return super().formatException(ei)
        return f"{ei[0].__name__}: {ei[1]!s}"



from __future__ import annotations

import datetime
import io
import time
from typing import TYPE_CHECKING

from monty.json import jsanitize
from rich.markdown import Markdown
from rich.panel import Panel
from rich.scope import render_scope
from rich.table import Table
from rich.text import Text

from jobflow_remote.cli.utils import ReprStr, fmt_datetime
from jobflow_remote.jobs.state import FlowState, JobState
from jobflow_remote.remote.data import get_job_path
from jobflow_remote.utils.data import convert_utc_time

if TYPE_CHECKING:
    from rich.console import RenderableType

    from jobflow_remote.config.base import ExecutionConfig, WorkerBase
    from jobflow_remote.jobs.data import FlowInfo, JobDoc, JobInfo
    from jobflow_remote.jobs.report import FlowsReport, JobsReport
    from jobflow_remote.jobs.upgrade import UpgradeAction


def format_state(ji: JobInfo) -> Text:
    state = ji.state.name
    if ji.state in (JobState.REMOTE_ERROR, JobState.FAILED):
        state = f"[bold red]{state}[/]"
    elif ji.remote.retry_time_limit is not None:
        state = f"[bold orange3]{state}[/]"
    return Text.from_markup(state)


def format_run_time(ji: JobInfo) -> str:
    prefix = ""
    if ji.state == JobState.RUNNING:
        run_time = ji.estimated_run_time
        prefix = "~"
    else:
        run_time = ji.run_time
    if not run_time:
        return ""
    m, s = divmod(run_time, 60)
    h, m = divmod(m, 60)
    return prefix + f"{h:g}:{m:02g}"


time_zone_str = f" [{time.tzname[0]}]"
header_name_data_getter_map = {
    "db_id": ("DB id", lambda ji: str(ji.db_id)),
    "name": ("Name", lambda ji: ji.name),
    "state": ("State", format_state),
    "job_id": ("Job id (Index)", lambda ji: f"{ji.uuid}  ({ji.index})"),
    "worker": ("Worker", lambda ji: ji.worker),
    "last_updated": (
        "Last updated" + time_zone_str,
        lambda ji: convert_utc_time(ji.updated_on).strftime(fmt_datetime),
    ),
    "queue_id": ("Queue id", lambda ji: ji.remote.process_id),
    "run_time": (Text("Run time [h:mm]", no_wrap=True), format_run_time),
    "retry_time": (
        "Retry time" + time_zone_str,
        lambda ji: convert_utc_time(ji.remote.retry_time_limit).strftime(fmt_datetime)
        if ji.remote.retry_time_limit
        else None,
    ),
    "prev_state": (
        "Prev state",
        lambda ji: ji.previous_state.name if ji.previous_state else None,
    ),
    "locked": ("Locked", lambda ji: "*" if ji.lock_id is not None else None),
    "lock_id": ("Lock id", lambda ji: str(ji.lock_id)),
    "lock_time": (
        "Lock time" + time_zone_str,
        lambda ji: convert_utc_time(ji.lock_time).strftime(fmt_datetime)
        if ji.lock_time
        else None,
    ),
}


def get_job_info_table(
    jobs_info: list[JobInfo],
    verbosity: int,
    output_keys: list[str] | None = None,
    stored_data_keys: list[str] | None = None,
) -> Table:
    stored_data_keys = stored_data_keys or []
    if not output_keys or verbosity > 0:
        all_output_keys = list(header_name_data_getter_map)
        output_keys = all_output_keys[:6]
        if verbosity >= 1:
            output_keys += all_output_keys[6:10]
        if verbosity == 1:
            output_keys.append(all_output_keys[10])
        if verbosity >= 2:
            output_keys += all_output_keys[11:13]
    all_display_keys = output_keys + stored_data_keys

    sdk_map = {
        k: (k, lambda x, k=k: x.stored_data.get(k) if x.stored_data else None)
        for k in stored_data_keys
    }
    full_map = header_name_data_getter_map | sdk_map

    table = Table(title="Jobs info")
    for key in all_display_keys:
        table.add_column(full_map[key][0])

    for ji in jobs_info:
        table.add_row(*(full_map[key][1](ji) for key in all_display_keys))

    return table


def get_flow_info_table(flows_info: list[FlowInfo], verbosity: int) -> Table:
    time_zone_str = f" [{time.tzname[0]}]"

    table = Table(title="Flows info")
    table.add_column("DB id")
    table.add_column("Name")
    table.add_column("State")
    table.add_column("Flow id")
    table.add_column("Num Jobs")
    table.add_column("Last updated" + time_zone_str)

    if verbosity >= 1:
        table.add_column("Workers")

        table.add_column("Job states")

    for fi in flows_info:
        # show the smallest Job db_id as db_id
        db_id = min(fi.db_ids)

        row = [
            str(db_id),
            fi.name,
            fi.state.name,
            fi.flow_id,
            str(len(fi.job_ids)),
            convert_utc_time(fi.updated_on).strftime(fmt_datetime),
        ]

        if verbosity >= 1:
            workers = set(fi.workers)
            row.append(", ".join(workers))
            job_states = "-".join(js.short_value for js in fi.job_states)
            row.append(job_states)

        table.add_row(*row)

    return table


JOB_INFO_ORDER = [
    "db_id",
    "uuid",
    "index",
    "name",
    "state",
    "error",
    "remote",
    "previous_state",
    "job",
    "created_on",
    "updated_on",
    "start_time",
    "end_time",
    "metadata",
    "run_dir",
    "parents",
    "priority",
    "worker",
    "resources",
    "exec_config",
    "lock_id",
    "lock_time",
    "stored_data",
]


def format_job_info(
    job_info: JobInfo | JobDoc, verbosity: int, show_none: bool = False
):
    d = job_info.dict(exclude_none=not show_none)
    if verbosity == 1:
        d.pop("job", None)

    # convert dates at the first level and for the remote error
    for k, v in d.items():
        if isinstance(v, datetime.datetime):
            d[k] = convert_utc_time(v).strftime(fmt_datetime)

    if d["remote"].get("retry_time_limit"):
        d["remote"]["retry_time_limit"] = convert_utc_time(
            d["remote"]["retry_time_limit"]
        ).strftime(fmt_datetime)

    d = jsanitize(d, allow_bson=True, enum_values=True, strict=True)
    error = d.get("error")
    if error:
        d["error"] = ReprStr(error)

    remote_error = d["remote"].get("error")
    if remote_error:
        d["remote"]["error"] = ReprStr(remote_error)

    if verbosity == 0:
        d["remote"].pop("queue_out", None)
        d["remote"].pop("queue_err", None)
    else:
        queue_out = d["remote"].get("queue_out")
        if queue_out:
            d["remote"]["queue_out"] = ReprStr(queue_out)
        queue_err = d["remote"].get("queue_err")
        if queue_err:
            d["remote"]["queue_err"] = ReprStr(queue_err)

    # reorder the keys
    # Do not check here that all the keys in JobInfo are in JOB_INFO_ORDER. Check in the tests
    sorted_d = {}
    for k in JOB_INFO_ORDER:
        if k in d:
            sorted_d[k] = d[k]

    return render_scope(sorted_d, sort_keys=False)


def format_flow_info(flow_info: FlowInfo) -> Table:
    title = f"Flow: {flow_info.name} - {flow_info.flow_id} - {flow_info.state.name}"
    table = Table(title=title)
    table.title_style = "bold"
    table.add_column("DB id")
    table.add_column("Name")
    table.add_column("State")
    table.add_column("Job id  (Index)")
    table.add_column("Worker")

    for i, job_id in enumerate(flow_info.job_ids):
        state = flow_info.job_states[i].name

        row = [
            str(flow_info.db_ids[i]),
            flow_info.job_names[i],
            state,
            f"{job_id}  ({flow_info.job_indexes[i]})",
            flow_info.workers[i],
        ]

        table.add_row(*row)

    return table


def get_exec_config_table(
    exec_config: dict[str, ExecutionConfig], verbosity: int = 0
) -> Table:
    table = Table(title="Execution config", show_lines=verbosity > 0)
    table.add_column("Name")
    if verbosity > 0:
        table.add_column("modules")
        table.add_column("export")
        table.add_column("pre_run")
        table.add_column("post_run")
    for name in sorted(exec_config):
        row = [Text(name, style="bold")]
        if verbosity > 0:
            ec = exec_config[name]
            from ruamel.yaml import YAML

            yaml = YAML()
            # The following should already be the case but we keep it to be sure
            yaml.default_flow_style = False
            if ec.modules:
                ec_modules_strio = io.StringIO()
                yaml.dump(ec.modules, ec_modules_strio)
                row.append(ec_modules_strio.getvalue())
            else:
                row.append("")
            if ec.export:
                ec_export_strio = io.StringIO()
                yaml.dump(ec.export, ec_export_strio)
                row.append(ec_export_strio.getvalue())
            else:
                row.append("")
            if ec.post_run:
                row.append(ec.post_run)
            else:
                row.append("")

        table.add_row(*row)

    return table


def get_worker_table(workers: dict[str, WorkerBase], verbosity: int = 0) -> Table:
    table = Table(title="Workers", show_lines=verbosity > 1)
    table.add_column("Name")
    if verbosity > 0:
        table.add_column("type")
    if verbosity == 1:
        table.add_column("info")
    elif verbosity > 1:
        table.add_column("details")

    for name in sorted(workers):
        row = [Text(name, style="bold")]
        worker = workers[name]
        if verbosity > 0:
            row.append(worker.type)
        if verbosity == 1:
            row.append(render_scope(worker.cli_info))
        elif verbosity > 1:
            d = worker.dict(exclude_none=True)
            d = jsanitize(d, allow_bson=False, enum_values=True)
            row.append(render_scope(d))

        table.add_row(*row)

    return table


def create_bar(count, max_count, size=30, color="white"):
    """Creates a text-based bar for a histogram with fixed color per state."""
    bar_filled = "█" * int(size * count / max_count)
    bar_empty = " " * (size - len(bar_filled))
    return f"[{color}]{bar_filled}[white]{bar_empty}"


def get_job_report_components(report: JobsReport) -> list[RenderableType]:
    components = []

    # Summary of Key Metrics
    summary_table = Table(title="Job Summary", title_style="bold green")
    summary_table.add_column("Metric", style="cyan", justify="right")
    summary_table.add_column("Count", style="green", justify="center")

    summary_table.add_row("Completed Jobs", str(report.completed))
    summary_table.add_row("Running Jobs", str(report.running))
    summary_table.add_row("Error Jobs", str(report.error))
    summary_table.add_row("Active Jobs", str(report.active))

    components.append(summary_table)

    # Job State Distribution
    components.append(
        Panel("[bold green]Job State Distribution[/bold green]", expand=False)
    )

    # Remove COMPLETED, as this will likely account for most of the jobs present in the DB
    state_counts = dict(report.state_counts)
    state_counts.pop(JobState.COMPLETED)

    # Find the max count to normalize the histograms
    max_count = max(*state_counts.values(), 1)

    total_count = sum(state_counts.values()) or 1

    # Display job states in a histogram
    state_colors = {
        JobState.WAITING: "grey39",
        JobState.READY: "cyan",
        JobState.CHECKED_OUT: "bright_cyan",
        JobState.UPLOADED: "deep_sky_blue1",
        JobState.SUBMITTED: "blue",
        JobState.RUNNING: "green",
        JobState.TERMINATED: "red",
        JobState.DOWNLOADED: "blue_violet",
        JobState.REMOTE_ERROR: "yellow",
        JobState.COMPLETED: "green",
        JobState.FAILED: "red",
        JobState.PAUSED: "magenta",
        JobState.STOPPED: "dark_orange",
        JobState.USER_STOPPED: "orange4",
        JobState.BATCH_SUBMITTED: "light_slate_blue",
        JobState.BATCH_RUNNING: "chartreuse3",
    }

    newline = ""
    for state, color in state_colors.items():
        if state not in state_counts:
            continue
        count = state_counts[state]
        percentage = round((count / total_count) * 100)
        bar = create_bar(count, max_count, color=color)
        components.extend(
            [f"{newline}{state.name:15} [{count:>3}] ({percentage:>3}%):", bar]
        )
        newline = "\n"

    # Longest Running Jobs
    if report.longest_running:
        longest_running_table = get_job_info_table(report.longest_running, verbosity=1)
        longest_running_table.title = "Longest running jobs"
        longest_running_table.title_style = "bold green"
        components.append(longest_running_table)

    # Worker Utilization
    if report.worker_utilization:
        worker_table = Table(title="Worker Jobs Distribution", title_style="bold green")
        worker_table.add_column("Worker", style="cyan", justify="center")
        worker_table.add_column("Job Count", style="green", justify="center")

        for worker, count in report.worker_utilization.items():
            worker_table.add_row(worker, str(count))

        components.append(worker_table)

    # Job Trends
    if report.trends:
        trends = report.trends
        trends_table = Table(
            title=f"Job Trends ({trends.num_intervals} {trends.interval}) [{trends.timezone}]",
            title_style="bold green",
        )
        trends_table.add_column("Date", justify="center", style="cyan", no_wrap=True)
        trends_table.add_column("Completed", justify="center", style="green")
        trends_table.add_column("Failed", justify="center", style="red")
        trends_table.add_column("Remote Error", justify="center", style="yellow")

        for i in range(trends.num_intervals):
            trends_table.add_row(
                trends.dates[i],
                f"{trends.completed[i]}",
                f"{trends.failed[i]}",
                f"{trends.remote_error[i]}",
            )

        components.append(trends_table)

    return components


def get_flow_report_components(report: FlowsReport) -> list[RenderableType]:
    components = []

    # Summary of Key Metrics
    summary_table = Table(title="Flow Summary", title_style="bold green")
    summary_table.add_column("Metric", style="cyan", justify="right")
    summary_table.add_column("Count", style="green", justify="center")

    summary_table.add_row("Completed Flows", str(report.completed))
    summary_table.add_row("Running Flows", str(report.running))
    summary_table.add_row("Error Flows", str(report.error))

    components.append(summary_table)

    # Job State Distribution
    components.append(
        Panel("[bold green]Flow State Distribution[/bold green]", expand=False)
    )

    # Remove COMPLETED, as this will likely account for most of the jobs present in the DB
    state_counts = dict(report.state_counts)
    state_counts.pop(FlowState.COMPLETED)

    # Find the max count to normalize the histograms
    max_count = max(*state_counts.values(), 1)

    total_count = sum(state_counts.values()) or 1

    # Display job states in a histogram
    state_colors = {
        FlowState.WAITING: "grey39",
        FlowState.READY: "cyan",
        FlowState.RUNNING: "green",
        FlowState.COMPLETED: "green",
        FlowState.FAILED: "red",
        FlowState.PAUSED: "magenta",
        FlowState.STOPPED: "dark_orange",
    }

    newline = ""
    for state, color in state_colors.items():
        if state not in state_counts:
            continue
        count = state_counts[state]
        percentage = round((count / total_count) * 100)
        bar = create_bar(count, max_count, color=color)
        components.extend(
            [f"{newline}{state.name:15} [{count:>3}] ({percentage:>3}%):", bar]
        )
        newline = "\n"

    # Job Trends
    if report.trends:
        trends = report.trends
        trends_table = Table(
            title=f"Flow Trends ({trends.num_intervals} {trends.interval}) [{trends.timezone}]",
            title_style="bold green",
        )
        trends_table.add_column("Date", justify="center", style="cyan", no_wrap=True)
        trends_table.add_column("Completed", justify="center", style="green")
        trends_table.add_column("Failed", justify="center", style="red")

        for i in range(trends.num_intervals):
            trends_table.add_row(
                trends.dates[i],
                f"{trends.completed[i]}",
                f"{trends.failed[i]}",
            )

        components.append(trends_table)

    return components


def format_upgrade_actions(actions: list[UpgradeAction]):
    msg = ""
    for action in actions:
        msg += f"* {action.description}\n"

    return Markdown(msg)


def get_batch_processes_table(
    batch_processes: dict,
    workers: dict[str, WorkerBase],
    running_jobs: dict[str, list[tuple[str, int, str]]],
    verbosity: int = 0,
):
    table = Table(title="Flows info")
    table.add_column("Process ID")
    table.add_column("Process UUID")
    table.add_column("Worker")
    table.add_column("Process folder")
    if verbosity > 0:
        table.add_column("Running Job ids (Index)")

    for worker_name, processes_data in batch_processes.items():
        worker = workers[worker_name]
        for process_id, process_uuid in processes_data.items():
            row = [
                process_id,
                process_uuid,
                worker_name,
                get_job_path(process_uuid, None, worker.batch.work_dir),
            ]

            if verbosity > 0:
                jobs_data = running_jobs.get(worker_name, [])
                process_jobs = [
                    f"{job_data[0]} ({job_data[1]})"
                    for job_data in jobs_data
                    if job_data[2] == process_uuid
                ]

                row.append("\n".join(process_jobs))

            table.add_row(*row)

    return table



# Import the submodules with a local app to register them to the main app
import jobflow_remote.cli.admin
import jobflow_remote.cli.backup
import jobflow_remote.cli.batch
import jobflow_remote.cli.execution
import jobflow_remote.cli.flow
import jobflow_remote.cli.gui
import jobflow_remote.cli.job
import jobflow_remote.cli.project
import jobflow_remote.cli.runner
from jobflow_remote.cli.jf import app



from __future__ import annotations

import logging
import os
import random
from contextlib import ExitStack
from pathlib import Path
from typing import TYPE_CHECKING

from flufl.lock import Lock, LockError

if TYPE_CHECKING:
    from jobflow_remote.remote.host import BaseHost


logger = logging.getLogger(__name__)


LOCK_DIR = "lock"

TERMINATED_DIR = "terminated"

RUNNING_DIR = "running"

SUBMITTED_DIR = "submitted"


class RemoteBatchManager:
    """
    Manager of remote files containing information about Jobs to be handled by
    a batch worker.

    Used by the Runner.
    """

    def __init__(
        self,
        host: BaseHost,
        files_dir: str | Path,
    ) -> None:
        """

        Parameters
        ----------
        host
            The host where the files are.
        files_dir
            The full path to directory where the files are stored.
        """
        self.host = host
        self.files_dir = Path(files_dir)
        self.submitted_dir = self.files_dir / SUBMITTED_DIR
        self.running_dir = self.files_dir / RUNNING_DIR
        self.terminated_dir = self.files_dir / TERMINATED_DIR
        self.lock_dir = self.files_dir / LOCK_DIR
        # All the directories need to be initialized to check that they exist
        # and the host connected.
        # Doing it here has two downsides: 1) it slows down the
        # start of the runner just for a batch worker being present
        # 2) if the connection cannot be established the runner may not
        # even start due to the connection errors.
        # The initialization is thus done once when the first action is performed.
        self._dir_initialized = False

    def _init_files_dir(self) -> None:
        """Initialize the file directory, creating all the subdirectories."""
        self.host.connect()
        self.host.mkdir(self.files_dir)
        self.host.mkdir(self.submitted_dir)
        self.host.mkdir(self.running_dir)
        self.host.mkdir(self.terminated_dir)
        self.host.mkdir(self.lock_dir)
        self._dir_initialized = True

    def submit_job(self, job_id: str, index: int) -> None:
        """
        Submit a Job by uploading the corresponding file.

        Parameters
        ----------
        job_id
            Uuid of the Job.
        index
            Index of the Job.
        """
        if not self._dir_initialized:
            self._init_files_dir()
        self.host.write_text_file(self.submitted_dir / f"{job_id}_{index}", "")

    def get_submitted(self) -> list[str]:
        """
        Get a list of files present in the submitted directory.

        Returns
        -------
            The list of file names in the submitted directory.
        """
        if not self._dir_initialized:
            self._init_files_dir()
        return self.host.listdir(self.submitted_dir)

    def get_terminated(self) -> list[tuple[str, int, str]]:
        """
        Get job ids and process ids of the terminated jobs from the corresponding
        directory on the host.

        Returns
        -------
        list
            The list of job ids, job indexes and batch process uuids in the host
            terminated directory.
        """
        if not self._dir_initialized:
            self._init_files_dir()
        terminated = []
        for i in self.host.listdir(self.terminated_dir):
            job_id, _index, process_uuid = i.split("_")
            index = int(_index)
            terminated.append((job_id, index, process_uuid))
        return terminated

    def get_running(self) -> list[tuple[str, int, str]]:
        """
        Get job ids and process ids of the running jobs from the corresponding
        directory on the host.

        Returns
        -------
        list
            The list of job ids, job indexes and batch process uuids in the host
            running directory.
        """
        if not self._dir_initialized:
            self._init_files_dir()
        running = []
        for filename in self.host.listdir(self.running_dir):
            job_id, _index, process_uuid = filename.split("_")
            index = int(_index)
            running.append((job_id, index, process_uuid))
        return running

    def delete_terminated(self, ids: list[tuple[str, int, str]]) -> None:
        if not self._dir_initialized:
            self._init_files_dir()
        for job_id, index, process_uuid in ids:
            self.host.remove(self.terminated_dir / f"{job_id}_{index}_{process_uuid}")


class LocalBatchManager:
    """
    Manager of local files  containing information about Jobs to be handled by
    a batch worker.

    Used in the worker to executes the batch Jobs.
    """

    def __init__(
        self,
        files_dir: str | Path,
        process_id: str,
        multiprocess_lock=None,
    ) -> None:
        """
        Parameters
        ----------
        files_dir
            The full path to directory where the files to handle the jobs
            to be executed in batch processes are stored.
        process_id
            The uuid associated to the batch process.
        multiprocess_lock
            A lock from the multiprocessing module to be used when executing jobs in
            parallel with other processes of the same worker.
        """
        self.process_id = process_id
        self.files_dir = Path(files_dir)
        self.multiprocess_lock = multiprocess_lock
        self.submitted_dir = self.files_dir / SUBMITTED_DIR
        self.running_dir = self.files_dir / RUNNING_DIR
        self.terminated_dir = self.files_dir / TERMINATED_DIR
        self.lock_dir = self.files_dir / LOCK_DIR

    def get_job(self) -> str | None:
        """
        Select randomly a job from the submitted directory to be executed.
        Move the file to the running directory.

        Locks will prevent the same job from being executed from other processes.
        If no job can be executed, None is returned.

        Returns
        -------
        str | None
            The name of the job that was selected, or None if no job can be executed.
        """
        files = os.listdir(self.submitted_dir)

        while files:
            selected = random.choice(files)
            try:
                with ExitStack() as lock_stack:
                    # if in a multiprocess execution, avoid concurrent interaction
                    # from processes belonging to the same job
                    if self.multiprocess_lock:
                        lock_stack.enter_context(self.multiprocess_lock)
                    lock_stack.enter_context(
                        Lock(
                            str(self.lock_dir / selected),
                            lifetime=60,
                            default_timeout=0,
                        )
                    )
                    os.remove(self.submitted_dir / selected)
                    (self.running_dir / f"{selected}_{self.process_id}").touch()
                    return selected
            except (LockError, FileNotFoundError):
                logger.exception(
                    f"Error while locking file {selected}. Will be ignored"
                )
                files.remove(selected)
        return None

    def terminate_job(self, job_id: str, index: int) -> None:
        """
        Terminate a job by removing the corresponding file from the running
        directory and adding a new file in the terminated directory.

        Parameters
        ----------
        job_id
            The uuid of the job to terminate.
        index
            The index of the job to terminate.
        """
        os.remove(self.running_dir / f"{job_id}_{index}_{self.process_id}")
        (self.terminated_dir / f"{job_id}_{index}_{self.process_id}").touch()



from typing import Annotated

import typer
from rich.text import Text

from jobflow_remote.cli.jfr_typer import JFRTyper
from jobflow_remote.cli.types import tree_opt
from jobflow_remote.cli.utils import (
    cleanup_job_controller,
    complete_profiling,
    get_config_manager,
    initialize_config_manager,
    out_console,
    start_profiling,
)
from jobflow_remote.config import ConfigError
from jobflow_remote.utils.log import initialize_cli_logger

app = JFRTyper(
    name="jf",
    add_completion=False,
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    epilog=None,  # to remove the default message in JFRTyper
)


def main_result_callback(*args, **kwargs) -> None:
    """
    Callback executed after the main command is completed.
    Allowing to make cleanup and other final actions.
    """
    cleanup_job_controller()
    profile = kwargs.get("profile", False)
    if profile:
        complete_profiling()


@app.callback(result_callback=main_result_callback)
def main(
    project: Annotated[
        str,
        typer.Option(
            "--project",
            "-p",
            help="Select a project for the current execution",
            is_eager=True,
        ),
    ] = None,
    full_exc: Annotated[
        bool,
        typer.Option(
            "--full-exc",
            "-fe",
            help="Print the full stack trace of exception when enabled",
            is_eager=True,
        ),
    ] = False,
    profile: Annotated[
        bool,
        typer.Option(
            "--profile",
            "-prof",
            help="Profile the command execution and provide a report at the end. For developers",
            is_eager=True,
            hidden=True,
        ),
    ] = False,
    print_tree: tree_opt = False,  # If selected will print the tree of the CLI and exit
) -> None:
    """The controller CLI for jobflow-remote."""
    from jobflow_remote import SETTINGS

    if full_exc:
        SETTINGS.cli_full_exc = True

    if profile:
        start_profiling()

    initialize_cli_logger(
        level=SETTINGS.cli_log_level.to_logging(), full_exc_info=SETTINGS.cli_full_exc
    )

    # initialize the ConfigManager only once, to avoid parsing the configuration
    # files multiple times when the command is executed.
    initialize_config_manager()
    cm = get_config_manager()
    if project:
        SETTINGS.project = project

    try:
        project_data = cm.get_project_data()
        text = Text.from_markup(
            f"The selected project is [green]{project_data.project.name}[/green] "
            f"from config file [green]{project_data.filepath}[/green]"
        )
        out_console.print(text)
    except ConfigError:
        # no warning printed if not needed as this seems to be confusing for the user
        pass



from typing import Annotated, Optional

import typer
from packaging.version import parse as parse_version
from rich.prompt import Confirm
from rich.text import Text

from jobflow_remote.cli.formatting import format_upgrade_actions
from jobflow_remote.cli.jf import app
from jobflow_remote.cli.jfr_typer import JFRTyper
from jobflow_remote.cli.types import (
    db_ids_opt,
    end_date_opt,
    flow_ids_opt,
    flow_state_opt,
    force_opt,
    foreground_index_opt,
    index_direction_arg,
    index_key_arg,
    job_ids_indexes_opt,
    job_state_opt,
    start_date_opt,
)
from jobflow_remote.cli.utils import (
    IndexDirection,
    check_stopped_runner,
    confirm_project_name,
    exit_with_error_msg,
    exit_with_warning_msg,
    get_config_manager,
    get_job_controller,
    get_job_ids_indexes,
    loading_spinner,
    out_console,
)
from jobflow_remote.jobs.data import DbCollection
from jobflow_remote.jobs.upgrade import DatabaseUpgrader

app_admin = JFRTyper(
    name="admin", help="Commands for administering the database", no_args_is_help=True
)
app.add_typer(app_admin)


@app_admin.command()
def upgrade(
    no_dry_run: Annotated[
        bool,
        typer.Option(
            "--no-dry-run",
            "-ndr",
            help="Skip the dry-run step",
        ),
    ] = False,
    check_env: Annotated[
        bool,
        typer.Option(
            "--check-env",
            "-ce",
            help="Report the changes in the package installed in the python environment before "
            "proceeding with the upgrade",
        ),
    ] = False,
    target: Annotated[
        Optional[str],
        typer.Option(
            "--target",
            "-t",
            help="Explicitly sets the target version for upgrade. For testing purposes or development"
            " versions. Not for standard updates.",
        ),
    ] = None,
) -> None:
    """
    Upgrade the jobflow database.
    WARNING: can modify all the data. Previous version cannot be retrieved anymore.
    It preferable to perform a backup before upgrading.
    """
    check_stopped_runner(error=True)

    jc = get_job_controller()
    upgrader = DatabaseUpgrader(jc)
    target_version = parse_version(target) if target else upgrader.current_version
    if target_version.local or target_version.post:
        # this is likely a developer version installed from source.
        # get the available upgrades for version higher than the developer one
        base_version = parse_version(target_version.base_version)
        upgrades_available = upgrader.collect_upgrades(
            base_version, upgrader.registered_upgrades[-1]
        )
        msg = (
            f"Target version {target_version} is likely a development version. Explicitly "
            f"specify the target version with the --target option if this is the case."
        )
        if upgrades_available:
            msg += (
                f" Available upgrades larger than {base_version}: "
                f"{','.join(str(v) for v in upgrades_available)}"
            )
        out_console.print(msg, style="gold1")
    db_version = jc.get_current_db_version()
    if db_version >= target_version:
        exit_with_warning_msg(
            f"Current DB version: {db_version}. No upgrade required for target version {target_version}"
        )

    jobflow_check = jc.upgrade_check_jobflow()
    if jobflow_check:
        out_console.print(jobflow_check)

    if check_env:
        full_check = jc.upgrade_full_check()
        if jobflow_check or full_check:
            text = Text.from_markup(
                +full_check
                + "This does not necessarily pose a problem. Just check if there is any potential "
                "incompatible upgrade"
            )
            out_console.print(text)

    if not no_dry_run:
        actions = upgrader.dry_run(target_version=target)
        if not actions:
            out_console.print(
                f"No actions will be required for upgrading to version {target_version}"
            )
        else:
            out_console.print(
                f"In order to upgrade the DB to version {target_version} the following actions will be performed:"
            )
            actions_formatted = format_upgrade_actions(actions)
            out_console.print(actions_formatted)

    warn_text = Text.from_markup(
        "[bold red]Performing the upgrade will permanently modify data in the queue DB. "
        "It is advisable to perform a backup before proceeding.[/]"
    )
    out_console.print(warn_text)
    confirm_project_name(style="bold red", exit_on_error=True)
    with loading_spinner(processing=False) as progress:
        progress.add_task(description="Upgrading the DB...", total=None)

        done = upgrader.upgrade(target_version=target)
    not_text = "" if done else "[bold]NOT [/bold]"
    out_console.print(f"The database has {not_text}been upgraded")


@app_admin.command()
def reset(
    validation: Annotated[
        Optional[str],
        typer.Argument(
            help=(
                "If the number of flows in the DB exceed 25 it will be required to pass "
                "today's date in the YYYY-MM-DD to proceed with the reset"
            ),
            metavar="DATE",
        ),
    ] = None,
    reset_output: Annotated[
        bool,
        typer.Option(
            "--output",
            "-o",
            help="Also delete all the documents in the current store",
        ),
    ] = False,
    force: force_opt = False,
) -> None:
    """
    Reset the jobflow database.
    WARNING: deletes all the data. These could not be retrieved anymore.
    """
    from jobflow_remote import SETTINGS

    check_stopped_runner(error=True)

    if not force:
        cm = get_config_manager()
        project_name = cm.get_project_data().project.name
        text = Text.from_markup(
            "[red]This operation will [bold]delete all the Jobs data[/bold] "
            f"for project [bold]{project_name}[/bold]. Proceed anyway?[/red]"
        )

        confirmed = Confirm.ask(text, default=False)
        if not confirmed:
            raise typer.Exit(0)
    with loading_spinner(processing=False) as progress:
        progress.add_task(description="Resetting the DB...", total=None)
        jc = get_job_controller()
        done = jc.reset(reset_output=reset_output, max_limit=25, validation=validation)
    not_text = "" if done else "[bold]NOT [/bold]"
    out_console.print(f"The database was {not_text}reset")
    if not done and SETTINGS.cli_suggestions:
        out_console.print(
            "Check the amount of Flows and set the DATE argument if this is the correct project to reset",
            style="yellow",
        )


@app_admin.command()
def unlock(
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    force: force_opt = False,
) -> None:
    """
    Forcibly removes the lock from the documents of the selected jobs.
    If no criteria is specified all the locked jobs will be selected.
    WARNING: can lead to inconsistencies if the processes is actually running.
    """
    job_ids_indexes = get_job_ids_indexes(job_id)

    jc = get_job_controller()

    if not force:
        with loading_spinner(processing=False) as progress:
            progress.add_task(
                description="Checking the number of locked documents...", total=None
            )

            jobs_info = jc.get_jobs_info(
                job_ids=job_ids_indexes,
                db_ids=db_id,
                states=state,
                start_date=start_date,
                locked=True,
                end_date=end_date,
            )

        if not jobs_info:
            exit_with_error_msg("No data matching the request")

        text = Text.from_markup(
            f"[red]This operation will [bold]remove the lock[/bold] for (roughly) [bold]{len(jobs_info)} Job(s)[/bold]. Proceed anyway?[/red]"
        )
        confirmed = Confirm.ask(text, default=False)

        if not confirmed:
            raise typer.Exit(0)

    with loading_spinner(processing=False) as progress:
        progress.add_task(description="Unlocking jobs...", total=None)

        num_unlocked = jc.unlock_jobs(
            job_ids=job_id,
            db_ids=db_id,
            states=state,
            start_date=start_date,
            end_date=end_date,
        )

    out_console.print(f"{num_unlocked} jobs were unlocked")


@app_admin.command()
def unlock_flow(
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: flow_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    force: force_opt = False,
) -> None:
    """
    Forcibly removes the lock from the documents of the selected jobs.
    If no criteria is specified all the locked flows will be selected.
    WARNING: can lead to inconsistencies if the processes is actually running.
    """
    job_ids_indexes = get_job_ids_indexes(job_id)

    jc = get_job_controller()

    if not force:
        with loading_spinner(processing=False) as progress:
            progress.add_task(
                description="Checking the number of locked documents...", total=None
            )

            flows_info = jc.get_flows_info(
                job_ids=job_ids_indexes,
                db_ids=db_id,
                flow_ids=flow_id,
                states=state,
                start_date=start_date,
                locked=True,
                end_date=end_date,
            )

        if not flows_info:
            exit_with_error_msg("No data matching the request")

        text = Text.from_markup(
            f"[red]This operation will [bold]remove the lock[/bold] for (roughly) [bold]{len(flows_info)} Flow(s)[/bold]. Proceed anyway?[/red]"
        )
        confirmed = Confirm.ask(text, default=False)

        if not confirmed:
            raise typer.Exit(0)

    with loading_spinner(processing=False) as progress:
        progress.add_task(description="Unlocking flows...", total=None)

        num_unlocked = jc.unlock_flows(
            job_ids=job_id,
            db_ids=db_id,
            flow_ids=flow_id,
            states=state,
            start_date=start_date,
            end_date=end_date,
        )

    out_console.print(f"{num_unlocked} flows were unlocked")


@app_admin.command()
def unlock_runner() -> None:
    """
    Forcibly removes the lock from the runner document.
    WARNING: can lead to inconsistencies if a runner daemon is actually running.
    """
    jc = get_job_controller()

    with loading_spinner(processing=False) as progress:
        progress.add_task(description="Unlocking runner document...", total=None)

        num_docs, num_unlocked = jc.unlock_runner()

    if num_docs == 0:
        out_console.print(
            "No runner document... "
            "Consider upgrading your database using 'jf admin upgrade'"
        )
    elif num_unlocked == 0:
        out_console.print("The runner document was not locked. Nothing changed.")
    else:
        out_console.print("The runner document has been unlocked.")


app_index = JFRTyper(
    name="index",
    help="Commands for managing the indexes of the queue database",
    no_args_is_help=True,
)
app_admin.add_typer(app_index)


@app_index.command()
def rebuild(
    foreground: foreground_index_opt = False,
):
    """
    Rebuild all the standard indexes.
    Old indexes will be dropped, including the custom ones.
    """
    with loading_spinner(processing=False) as progress:
        progress.add_task(description="Resetting the indexes...", total=None)
        jc = get_job_controller()
        jc.build_indexes(background=not foreground, drop=True)

    if foreground:
        out_console.print("Indexes rebuilt")
    else:
        out_console.print("Indexes rebuild started in background")


@app_index.command()
def create(
    key: index_key_arg,
    direction: index_direction_arg = IndexDirection.ASC,
    foreground: foreground_index_opt = False,
    collection: Annotated[
        Optional[DbCollection],
        typer.Option(
            "--collection",
            "-c",
            help="The collection where the index should be added",
        ),
    ] = DbCollection.JOBS.value,  # type: ignore[assignment]
    unique: Annotated[
        bool,
        typer.Option(
            "--unique",
            "-u",
            help="The index will be unique",
        ),
    ] = False,
):
    """
    Add an index to one of the queue collections
    """
    with loading_spinner(processing=False) as progress:
        progress.add_task(description="Resetting the indexes...", total=None)
        jc = get_job_controller()
        jc.create_indexes(
            indexes=[[(key, direction.as_pymongo)]],
            unique=unique,
            collection=collection,
            background=not foreground,
        )
    if foreground:
        out_console.print("Index created")
    else:
        out_console.print("Index creation started in background")



from __future__ import annotations

import contextlib
import functools
import logging
from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING, Callable, ClassVar

from packaging.version import Version
from packaging.version import parse as parse_version

import jobflow_remote

if TYPE_CHECKING:
    from pymongo.client_session import ClientSession

    from jobflow_remote.jobs.jobcontroller import JobController

logger = logging.getLogger(__name__)


class UpgradeRequiredError(Exception):
    """
    An error signaling that an upgrade should be performed before performing any further action.
    """


@dataclass
class UpgradeAction:
    """Details of a single upgrade action to be performed"""

    description: str
    collection: str
    action_type: str
    details: dict
    required: bool = False


class DatabaseUpgrader:
    """
    Object to handle the upgrade of the database between different versions
    """

    _upgrade_registry: ClassVar[dict[Version, Callable]] = {}

    def __init__(self, job_controller: JobController):
        self.job_controller = job_controller
        self.current_version = parse_version(jobflow_remote.__version__)

    @classmethod
    def register_upgrade(cls, version: str):
        """Decorator to register upgrade functions.

        This decorator should be used to register functions that implement the upgrades for each
        version.

        Parameters
        ----------
        version
            The version to register the upgrade function for
        """

        def decorator(func: Callable):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                logger.info(f"Executing upgrade to version {version}")
                start_time = datetime.now()
                result = func(*args, **kwargs)
                duration = (datetime.now() - start_time).total_seconds()
                logger.info(f"Completed upgrade to {version} in {duration:.2f}s")
                return result

            cls._upgrade_registry[parse_version(version)] = wrapper
            return wrapper

        return decorator

    @property
    def registered_upgrades(self):
        return sorted(self._upgrade_registry.keys())

    def collect_upgrades(
        self, from_version: Version, target_version: Version
    ) -> list[Version]:
        """
        Determines the upgrades that need to be performed.

        from_version
            The version using as a starting point for the list of upgrades.
        target_version
            The final version up to which the upgrades should be listed.
        """
        return [
            v for v in self.registered_upgrades if from_version < v <= target_version
        ]

    def update_db_version(self, version: Version, session: ClientSession | None = None):
        """
        Update the jobflow-remote version information stored in the database.

        Parameters
        ----------
        version
            The version to update the database to
        session
            The client session to use to perform the update
        """
        self.job_controller.auxiliary.update_one(
            {"jobflow_remote_version": {"$exists": True}},
            {"$set": {"jobflow_remote_version": str(version)}},
            upsert=True,
            session=session,
        )

    def dry_run(
        self, from_version: str | None = None, target_version: str | None = None
    ) -> list[UpgradeAction]:
        """Simulate the upgrade process and return all actions that would be performed

        Parameters
        ----------
        from_version
            The version from which to start the upgrade. If ``None``, the current version in the database is used.
        target_version
            The target version of the upgrade. If ``None``, the current version of the package is used.

        Returns
        -------
        list
            A list of UpgradeAction objects describing all actions that would be performed
            during the upgrade.
        """
        db_version = (
            parse_version(from_version)
            if from_version
            else self.job_controller.get_current_db_version()
        )
        target_version = (
            parse_version(target_version) if target_version else self.current_version
        )

        if db_version >= target_version:
            return []

        versions_needing_upgrade = self.collect_upgrades(db_version, target_version)

        all_actions = []
        for version in versions_needing_upgrade:
            upgrade_func = self._upgrade_registry[version]
            actions = upgrade_func(self.job_controller, dry_run=True)
            all_actions.extend(actions)

        # Add the version update action
        all_actions.append(
            UpgradeAction(
                description=f"Update database version number to {target_version}",
                collection="auxiliary",
                action_type="update",
                details={
                    "filter": {"jobflow_remote_version": {"$exists": True}},
                    "update": {"$set": {"jobflow_remote_version": str(target_version)}},
                    "upsert": True,
                },
                required=False,
            )
        )

        return all_actions

    def upgrade(
        self, from_version: str | None = None, target_version: str | None = None
    ) -> bool:
        """Perform the database upgrade

        This method will check if an upgrade is needed from the given version
        to the target version and execute the necessary upgrade functions.
        If no target version is provided, the current version of the package
        is used.

        Parameters
        ----------
        from_version
            The version from which to start the upgrade. If ``None``, the current version in the database is used.
        target_version
            The target version of the upgrade. If ``None``, the current version of the package is used.

        Returns
        -------
        bool
            True if the upgrade was performed.
        """
        db_version = (
            parse_version(from_version)
            if from_version
            else self.job_controller.get_current_db_version()
        )
        target_version = (
            parse_version(target_version) if target_version else self.current_version
        )
        if db_version >= target_version:
            logger.info("Database is already at the target version")
            return False

        versions_needing_upgrade = self.collect_upgrades(db_version, target_version)

        logger.info(f"Starting upgrade from version {db_version} to {target_version}")

        for version in versions_needing_upgrade:
            with self.open_transaction() as session:
                upgrade_func = self._upgrade_registry[version]
                logger.info(f"Applying upgrade to version {version}")
                upgrade_func(self.job_controller, session=session)
                self.update_db_version(version, session)

        # update the full environment reference and versions
        logger.info("Updating database information")
        self.job_controller.update_version_information(
            jobflow_remote_version=target_version
        )

        logger.info("Database upgrade completed successfully")
        return True

    @contextlib.contextmanager
    def open_transaction(self):
        """
        Open a transaction for the queue DB in the jobstore if it is supported.
        Does nothing and yields None if transactions are not supported
        """
        if self.job_controller.queue_supports_transactions:
            with (
                self.job_controller.db.client.start_session() as session,
                session.start_transaction(),
            ):
                yield session
        else:
            yield None


@DatabaseUpgrader.register_upgrade("0.1.5")
def upgrade_to_0_1_5(
    job_controller: JobController,
    session: ClientSession | None = None,
    dry_run: bool = False,
) -> list[UpgradeAction]:
    actions = []
    action = UpgradeAction(
        description="Create a document for the running runner in the auxiliary collection",
        collection="auxiliary",
        action_type="update",
        details={
            "filter": {"running_runner": {"$exists": True}},
            "update": {"$set": {"running_runner": None}},
            "upsert": True,
            "required": True,
        },
    )

    if not dry_run:
        job_controller.auxiliary.find_one_and_update(
            filter=action.details["filter"],
            update=action.details["update"],
            upsert=action.details["upsert"],
            session=session,
        )

    actions.append(action)
    return actions



from __future__ import annotations

from typing import TYPE_CHECKING

from jobflow_remote.config.manager import ConfigManager

if TYPE_CHECKING:
    from jobflow.core.store import JobStore


def get_jobstore(project_name: str | None = None) -> JobStore:
    """
    Helper function to get the jobstore in a project.

    Parameters
    ----------
    project_name
        Name of the project or None to use the one from the settings.

    Returns
    -------
    A JobStore
    """
    cm = ConfigManager(warn=False)
    project = cm.get_project(project_name=project_name)
    return project.get_jobstore()



from __future__ import annotations

import re
from typing import IO, TYPE_CHECKING, Any

from typer.testing import CliRunner, Result

from jobflow_remote.cli.jf import app

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence


def run_check_cli(
    cli_args: str | Sequence[str] | None = None,
    cli_input: bytes | str | IO[Any] | None = None,
    cli_env: Mapping[str, str] | None = None,
    catch_exceptions: bool = False,
    required_out: str | Sequence[str] | None = None,
    excluded_out: str | Sequence[str] | None = None,
    error: bool = False,
    terminal_width: int = 1000,
) -> Result:
    if isinstance(required_out, str):
        required_out = [required_out]
    if isinstance(excluded_out, str):
        excluded_out = [excluded_out]

    cli_runner = CliRunner()

    result = cli_runner.invoke(
        app,
        args=cli_args,
        input=cli_input,
        env=cli_env,
        catch_exceptions=catch_exceptions,
        terminal_width=terminal_width,
    )

    # note that stderr is not captured separately
    assert (
        error == (result.exit_code != 0)
    ), f"cli should have {'' if error else 'not '}failed. exit code: {result.exit_code}. stdout: {result.stdout}"

    # the print of the output in the console during the tests may result in newlines added
    # that prevent the output to be matched. replace all spaces with a single space.
    single_space_output = re.sub(r"[\n\t\s]*", " ", result.stdout)

    if required_out:
        for ro in required_out:
            assert (
                re.sub(r"[\n\t\s]*", " ", ro) in single_space_output
            ), f"{ro} missing from stdout: {result.stdout}"

    if excluded_out:
        for eo in excluded_out:
            assert (
                re.sub(r"[\n\t\s]*", " ", eo) not in single_space_output
            ), f"{eo} present in stdout: {result.stdout}"

    return result



from typing import Annotated, Optional

import typer

from jobflow_remote.cli.formatting import get_batch_processes_table
from jobflow_remote.cli.jf import app
from jobflow_remote.cli.jfr_typer import JFRTyper
from jobflow_remote.cli.types import verbosity_opt
from jobflow_remote.cli.utils import (
    exit_with_warning_msg,
    get_config_manager,
    get_job_controller,
    out_console,
)
from jobflow_remote.jobs.batch import RemoteBatchManager

app_batch = JFRTyper(
    name="batch", help="Helper utils handling batch jobs", no_args_is_help=True
)
app.add_typer(app_batch)


@app_batch.command(name="list")
def processes_list(
    worker: Annotated[
        Optional[str],
        typer.Option(
            "--worker",
            "-w",
            help="Select the worker.",
        ),
    ] = None,
    verbosity: verbosity_opt = 0,
) -> None:
    """
    Show the list of processes being executed on the batch workers.
    Increasing verbosity will require connecting to the host.
    """

    jc = get_job_controller()

    batch_processes = jc.get_batch_processes(worker)
    if not batch_processes or not any(wbc for wbc in batch_processes.values()):
        exit_with_warning_msg("No batch processes running")

    cm = get_config_manager()
    project = cm.get_project()
    workers = project.workers
    running_jobs = {}
    if verbosity > 0:
        for worker_name in batch_processes:
            worker_config = workers[worker_name]
            host = worker_config.get_host()
            host.connect()
            remote_batch_manager = RemoteBatchManager(
                host, worker_config.batch.jobs_handle_dir
            )
            running_jobs[worker_name] = remote_batch_manager.get_running()

    table = get_batch_processes_table(
        batch_processes=batch_processes,
        workers=workers,
        running_jobs=running_jobs,
        verbosity=verbosity,
    )

    out_console.print(table)



from collections import defaultdict
from datetime import datetime
from enum import Enum
from functools import cached_property
from typing import Optional, Union

from jobflow import Flow, Job
from monty.json import jsanitize
from pydantic import BaseModel, Field
from qtoolkit.core.data_objects import QResources, QState

from jobflow_remote.config.base import ExecutionConfig
from jobflow_remote.jobs.state import FlowState, JobState

IN_FILENAME = "jfremote_in.json"
OUT_FILENAME = "jfremote_out.json"


def get_initial_job_doc_dict(
    job: Job,
    parents: Optional[list[str]],
    db_id: str,
    worker: str,
    exec_config: Optional[ExecutionConfig],
    resources: Optional[Union[dict, QResources]],
    priority: int,
) -> dict:
    """
    Generate an instance of JobDoc for initial insertion in the DB.

    Parameters
    ----------
    job:
        The Job of the JobDoc.
    parents
        The parents of the Job.
    db_id
        The db_id.
    worker
        The worker where the Job should be executed.
    exec_config
        The ExecutionConfig used for execution.
    resources
        The resources used to run the Job.
    priority
        The priority of the Job.

    Returns
    -------
    JobDoc
        A new JobDoc.
    """
    from monty.json import jsanitize

    # take the resources either from the job, if they are defined
    # (they can be defined dynamically by the update_config) or the
    # defined value
    # Since  the resources can be dictionaries also allow to be an empty
    # dictionary.
    config_resources = job.config.manager_config.get("resources")
    job_resources = config_resources if config_resources is not None else resources
    job_exec_config = job.config.manager_config.get("exec_config") or exec_config
    worker = job.config.manager_config.get("worker") or worker
    priority = job.config.manager_config.get("priority") or priority

    job_doc = JobDoc(
        job=jsanitize(job, strict=True, enum_values=True),
        uuid=job.uuid,
        index=job.index,
        db_id=db_id,
        state=JobState.WAITING if parents else JobState.READY,
        parents=parents,
        worker=worker,
        exec_config=job_exec_config,
        resources=job_resources,
        priority=priority,
    )

    return job_doc.as_db_dict()


def get_initial_flow_doc_dict(flow: Flow, job_dicts: list[dict]) -> dict:
    """
    Generate a serialized FlowDoc for initial insertion in the DB.

    Parameters
    ----------
    flow
        The Flow used to generate the FlowDoc.
    job_dicts
        The dictionaries of the Jobs composing the Flow.

    Returns
    -------
    dict
        A serialized version of a new FlowDoc.
    """
    jobs = [j["uuid"] for j in job_dicts]
    ids = [(j["db_id"], j["uuid"], j["index"]) for j in job_dicts]
    parents = {j["uuid"]: {"1": j["parents"]} for j in job_dicts}

    flow_doc = FlowDoc(
        uuid=flow.uuid,
        jobs=jobs,
        state=FlowState.READY,
        name=flow.name,
        ids=ids,
        parents=parents,
        metadata=flow.metadata or {},
    )

    return flow_doc.as_db_dict()


class RemoteInfo(BaseModel):
    """Model with data describing the remote state of a Job."""

    step_attempts: int = 0
    queue_state: Optional[QState] = None
    process_id: Optional[str] = None
    retry_time_limit: Optional[datetime] = None
    error: Optional[str] = None
    prerun_cleanup: bool = False
    queue_out: Optional[str] = None
    queue_err: Optional[str] = None


class JobInfo(BaseModel):
    """
    Model with information extracted from a JobDoc.
    Mainly for visualization purposes.
    """

    uuid: str
    index: int
    db_id: str
    worker: str
    name: str
    state: JobState
    created_on: datetime
    updated_on: datetime
    remote: RemoteInfo = RemoteInfo()
    parents: Optional[list[str]] = None
    previous_state: Optional[JobState] = None
    error: Optional[str] = None
    lock_id: Optional[str] = None
    lock_time: Optional[datetime] = None
    run_dir: Optional[str] = None
    start_time: Optional[datetime] = None
    end_time: Optional[datetime] = None
    priority: int = 0
    metadata: Optional[dict] = None
    stored_data: Optional[dict] = None

    @property
    def is_locked(self) -> bool:
        return self.lock_id is not None

    @property
    def run_time(self) -> Optional[float]:
        """
        Calculate the run time based on start and end time.

        Returns
        -------
        float
            The run time in seconds
        """
        if self.start_time and self.end_time:
            return (self.end_time - self.start_time).total_seconds()

        return None

    @property
    def estimated_run_time(self) -> Optional[float]:
        """
        Estimate the current run time based on the start time and the current time.

        Returns
        -------
        float
            The estimated run time in seconds.
        """
        if self.start_time:
            return (datetime.utcnow() - self.start_time).total_seconds()

        return None

    @classmethod
    def from_query_output(cls, d) -> "JobInfo":
        """
        Generate an instance from the output of a query to the JobDoc collection.

        Parameters
        ----------
        d
            The dictionary with the queried data.

        Returns
        -------
        JobInfo
            The instance of JobInfo based on the data
        """
        job = d.pop("job")
        for k in ["name", "metadata"]:
            d[k] = job[k]
        return cls.model_validate(d)


def _projection_db_info() -> list[str]:
    """
    Generate a list of fields used for projection, depending on the JobInfo model.

    Returns
    -------
    list
        The list of fields to use in a query.
    """
    projection = list(JobInfo.model_fields)
    projection.remove("name")
    projection.append("job.name")
    projection.append("job.metadata")
    return projection


# generate the list only once.
projection_job_info = _projection_db_info()


class JobDoc(BaseModel):
    """Model for the standard representation of a Job in the queue database."""

    # TODO consider defining this as a dict and provide a get_job() method to
    # get the real Job. This would avoid (de)serializing jobs if this document
    # is used often to interact with the DB.
    job: Job
    uuid: str
    index: int
    db_id: str
    worker: str
    state: JobState
    remote: RemoteInfo = RemoteInfo()
    # only the uuid as list of parents for a JobDoc (i.e. uuid+index) is
    # enough to determine the parents, since once a job with a uuid is
    # among the parents, all the index will still be parents.
    # Note that for just the uuid this condition is not true: JobDocs with
    # the same uuid but different indexes may have different parents
    parents: Optional[list[str]] = None
    previous_state: Optional[JobState] = None
    error: Optional[str] = None  # TODO is there a better way to serialize it?
    lock_id: Optional[str] = None
    lock_time: Optional[datetime] = None
    run_dir: Optional[str] = None
    start_time: Optional[datetime] = None
    end_time: Optional[datetime] = None
    created_on: datetime = Field(default_factory=datetime.utcnow)
    updated_on: datetime = Field(default_factory=datetime.utcnow)
    priority: int = 0
    # store: Optional[JobStore] = None
    exec_config: Optional[Union[ExecutionConfig, str]] = None
    resources: Optional[Union[QResources, dict]] = None

    stored_data: Optional[dict] = None
    # history: Optional[list[str]] = None

    def as_db_dict(self) -> dict:
        """
        Generate a dict representation suitable to be inserted in the database.

        Returns
        -------
        dict
            The dict representing the JobDoc.
        """
        d = jsanitize(
            self.model_dump(mode="python"),
            strict=True,
            allow_bson=True,
            enum_values=True,
        )
        # required since the resources are not serialized otherwise
        if isinstance(self.resources, QResources):
            d["resources"] = self.resources.as_dict()
        return d


class FlowDoc(BaseModel):
    """Model for the standard representation of a Flow in the queue database."""

    uuid: str
    jobs: list[str]
    state: FlowState
    name: str
    lock_id: Optional[str] = None
    lock_time: Optional[datetime] = None
    created_on: datetime = Field(default_factory=datetime.utcnow)
    updated_on: datetime = Field(default_factory=datetime.utcnow)
    metadata: dict = Field(default_factory=dict)
    # parents need to include both the uuid and the index.
    # When dynamically replacing a Job with a Flow some new Jobs will
    # be parents of the job with index=i+1, but will not be parents of
    # the job with index i.
    # index is stored as string, since mongodb needs string keys
    # This dictionary include {job uuid: {job index: [parent's uuids]}}
    parents: dict[str, dict[str, list[str]]] = Field(default_factory=dict)
    # ids correspond to db_id, uuid, index for each JobDoc
    ids: list[tuple[str, str, int]] = Field(default_factory=list)

    def as_db_dict(self) -> dict:
        """
        Generate a dict representation suitable to be inserted in the database.

        Returns
        -------
        dict
            The dict representing the FlowDoc.
        """
        return jsanitize(
            self.model_dump(mode="python"),
            strict=True,
            allow_bson=True,
            enum_values=True,
        )

    @cached_property
    def int_index_parents(self):
        d = defaultdict(dict)
        for child_id, index_parents in self.parents.items():
            for index, parents in index_parents.items():
                d[child_id][int(index)] = parents
        return dict(d)

    @cached_property
    def children(self) -> dict[str, list[tuple[str, int]]]:
        d = defaultdict(list)
        for job_id, index_parents in self.parents.items():
            for index, parents in index_parents.items():
                for parent_id in parents:
                    d[parent_id].append((job_id, int(index)))

        return dict(d)

    def descendants(self, job_uuid: str) -> list[tuple[str, int]]:
        descendants = set()

        def add_descendants(uuid) -> None:
            children = self.children.get(uuid)
            if children:
                descendants.update(children)
                for child in children:
                    add_descendants(child[0])

        add_descendants(job_uuid)

        return list(descendants)

    @cached_property
    def ids_mapping(self) -> dict[str, dict[int, str]]:
        d: dict = defaultdict(dict)

        for db_id, job_id, index in self.ids:
            d[job_id][int(index)] = db_id

        return dict(d)


class RemoteError(RuntimeError):
    """An exception signaling errors during the update of the remote states."""

    def __init__(self, msg, no_retry=False) -> None:
        self.msg = msg
        self.no_retry = no_retry


projection_flow_info_jobs = [
    "db_id",
    "uuid",
    "index",
    "state",
    "job.name",
    "worker",
    "parents",
    "job.hosts",
]


class FlowInfo(BaseModel):
    """
    Model with information extracted from a FlowDoc.
    Mainly for visualization purposes.
    """

    db_ids: list[str]
    job_ids: list[str]
    job_indexes: list[int]
    flow_id: str
    state: FlowState
    name: str
    created_on: datetime
    updated_on: datetime
    workers: list[str]
    job_states: list[JobState]
    job_names: list[str]
    parents: list[list[str]]
    hosts: list[list[str]]

    @classmethod
    def from_query_dict(cls, d) -> "FlowInfo":
        created_on = d["created_on"]
        updated_on = d["updated_on"]
        flow_id = d["uuid"]
        jobs_data = d.get("jobs_list") or []

        workers = []
        job_states = []
        job_names = []
        parents = []
        job_hosts = []

        if jobs_data:
            db_ids = []
            job_ids = []
            job_indexes = []
            for job_doc in jobs_data:
                db_ids.append(job_doc["db_id"])
                job_ids.append(job_doc["uuid"])
                job_indexes.append(job_doc["index"])
                job_names.append(job_doc["job"]["name"])
                state = job_doc["state"]
                job_states.append(JobState(state))
                workers.append(job_doc["worker"])
                parents.append(job_doc["parents"] or [])
                job_hosts.append(job_doc["job"]["hosts"] or [])
        else:
            db_ids, job_ids, job_indexes = list(  # type:ignore[assignment]
                zip(*d["ids"])
            )
            # parents could be determined in this case as well from the Flow document.
            # However, to match the correct order it would require lopping over them.
            # To keep the generation faster add this only if a use case shows up.

        state = FlowState(d["state"])

        return cls(
            db_ids=db_ids,
            job_ids=job_ids,
            job_indexes=job_indexes,
            flow_id=flow_id,
            state=state,
            name=d["name"],
            created_on=created_on,
            updated_on=updated_on,
            workers=workers,
            job_states=job_states,
            job_names=job_names,
            parents=parents,
            hosts=job_hosts,
        )

    @cached_property
    def ids_mapping(self) -> dict[str, dict[int, str]]:
        d: dict = defaultdict(dict)

        for db_id, job_id, index in zip(self.db_ids, self.job_ids, self.job_indexes):
            d[job_id][int(index)] = db_id

        return dict(d)

    def iter_job_prop(self):
        n_jobs = len(self.job_ids)
        for i in range(n_jobs):
            d = {
                "db_id": self.db_ids[i],
                "uuid": self.job_ids[i],
                "index": self.job_indexes[i],
            }
            if self.job_names:
                d["name"] = self.job_names[i]
                d["state"] = self.job_states[i]
                d["parents"] = self.parents[i]
                d["hosts"] = self.hosts[i]
            yield d


class DynamicResponseType(Enum):
    """Types of dynamic responses in jobflow."""

    REPLACE = "replace"
    DETOUR = "detour"
    ADDITION = "addition"


class DbCollection(Enum):
    JOBS = "jobs"
    FLOWS = "flows"
    AUX = "aux"


def get_reset_job_base_dict() -> dict:
    """
    Generate a dictionary with the basic properties to update in case of reset.

    Returns
    -------
    dict
        Data to be reset.
    """
    return {
        "remote.step_attempts": 0,
        "remote.retry_time_limit": None,
        "previous_state": None,
        "remote.queue_state": None,
        "remote.error": None,
        "remote.queue_out": None,
        "remote.queue_err": None,
        "error": None,
        "updated_on": datetime.utcnow(),
        "start_time": None,
        "end_time": None,
    }



from __future__ import annotations

from typing import TYPE_CHECKING

from jobflow.core.flow import get_flow

from jobflow_remote.config.base import ConfigError, ExecutionConfig
from jobflow_remote.config.manager import ConfigManager
from jobflow_remote.remote.data import check_additional_stores

if TYPE_CHECKING:
    import jobflow
    from qtoolkit.core.data_objects import QResources


def submit_flow(
    flow: jobflow.Flow | jobflow.Job | list[jobflow.Job],
    worker: str | None = None,
    project: str | None = None,
    exec_config: str | ExecutionConfig | None = None,
    resources: dict | QResources | None = None,
    priority: int = 0,
    allow_external_references: bool = False,
) -> list[str]:
    """
    Submit a flow for calculation to the selected Worker.

    This will not start the calculation but just add to the database of the
    calculation to be executed.

    Parameters
    ----------
    flow
        A flow or job.
    worker
        The name of the Worker where the calculation will be submitted. If None, use the
        first configured worker for this project.
    project
        the name of the project to which the Flow should be submitted. If None the
        current project will be used.
    exec_config
        the options to set before the execution of the job in the submission script.
        In addition to those defined in the Worker.
    resources
        Information passed to qtoolkit to require the resources for the submission
        to the queue.
    priority
        The priority assigned to the Jobs.
    allow_external_references
        If False all the references to other outputs should be from other Jobs
        of the Flow.

    Returns
    -------
    List of str
        The list of db_ids of the submitted Jobs.
    """
    config_manager = ConfigManager()

    proj_obj = config_manager.get_project(project)
    if worker is None:
        if not proj_obj.workers:
            raise ConfigError("No workers configured for this project.")
        worker = next(iter(proj_obj.workers))

    # try to load the worker and exec_config to check that the values are well defined
    config_manager.get_worker(worker_name=worker, project_name=project)
    if isinstance(exec_config, str):
        config_manager.get_exec_config(
            exec_config_name=exec_config, project_name=project
        )

    flow = get_flow(flow, allow_external_references=allow_external_references)

    # check that all the additional stores are properly defined
    jobstore = proj_obj.get_jobstore()
    for job, _ in flow.iterflow():
        missing_stores = check_additional_stores(job, jobstore)
        if missing_stores:
            raise ConfigError(
                f"Additional stores {missing_stores!r} are not configured for this project."
            )

    jc = proj_obj.get_job_controller()

    return jc.add_flow(
        flow=flow,
        worker=worker,
        exec_config=exec_config,
        resources=resources,
        priority=priority,
        allow_external_references=allow_external_references,
    )



from __future__ import annotations

import abc
import logging
import re
import traceback
from typing import TYPE_CHECKING

from monty.json import MSONable

if TYPE_CHECKING:
    from pathlib import Path


logger = logging.getLogger(__name__)

SANITIZE_KEY = r"_-_-_-_-_### JFREMOTE SANITIZE ###_-_-_-_-_"


class BaseHost(MSONable):
    """Base Host class."""

    def __init__(self, sanitize: bool = False):
        """
        Parameters
        ----------
        sanitize
            If True text a string will be prepended and appended to the output
            of the commands, to ease the parsing and avoid failures due to spurious
            text coming from the host shell.
        """
        self.sanitize = sanitize
        self._sanitize_regex: re.Pattern | None = None

    @abc.abstractmethod
    def execute(
        self,
        command: str | list[str],
        workdir: str | Path | None = None,
        timeout: int | None = None,
    ) -> tuple[str, str, int]:
        """Execute the given command on the host.

        Parameters
        ----------
        command: str or list of str
            Command to execute, as a str or list of str
        workdir: str or None
            path where the command will be executed.
        timeout
            Timeout for the execution of the commands.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def mkdir(
        self, directory: str | Path, recursive: bool = True, exist_ok: bool = True
    ) -> bool:
        """Create directory on the host."""
        raise NotImplementedError

    @abc.abstractmethod
    def write_text_file(self, filepath, content):
        """Write content to a file on the host."""
        raise NotImplementedError

    @abc.abstractmethod
    def connect(self):
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> bool:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def is_connected(self) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def put(self, src, dst):
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, src, dst):
        raise NotImplementedError

    @abc.abstractmethod
    def copy(self, src, dst):
        raise NotImplementedError

    def test(self) -> str | None:
        msg = None
        try:
            cmd = "echo 'test'"
            stdout, stderr, returncode = self.execute(cmd)
            if returncode != 0:
                msg = f"Command was executed but return code was different from zero.\nstdoud: {stdout}\nstderr: {stderr}"
            elif stdout.strip() != "test" or stderr.strip() != "":
                msg = (
                    "Command was executed but the output is not the expected one (i.e. a single 'test' "
                    f"string in both stdout and stderr).\nstdoud: {stdout}\nstderr: {stderr}"
                )
                if not self.sanitize:
                    msg += (
                        "\nIf the output contains additional text the problem may be solved by setting "
                        "the 'sanitize_command' option to True in the project configuration."
                    )

        except Exception:
            exc = traceback.format_exc()
            msg = f"Error while executing command:\n {exc}"

        return msg

    @abc.abstractmethod
    def listdir(self, path: str | Path) -> list[str]:
        raise NotImplementedError

    @abc.abstractmethod
    def remove(self, path: str | Path):
        raise NotImplementedError

    @abc.abstractmethod
    def rmtree(self, path: str | Path, raise_on_error: bool = False) -> bool:
        """Recursively delete a directory tree on a host.

        This method must be implemented by subclasses of  `BaseHost`.
        It is intended to remove an entire directory tree, including all files
        and subdirectories, on the host represented by the subclass.

        Parameters
        ----------
        path : str or Path
            The path to the directory tree to be removed.

        raise_on_error : bool
            If set to `False` (default), errors will be ignored, and the method will
            attempt to continue removing remaining files and directories.
            Otherwise, any errors encountered during the removal process
            will raise an exception.

        Returns
        -------
        bool
            True if the directory tree was successfully removed, False otherwise.
        """
        raise NotImplementedError

    @property
    def interactive_login(self) -> bool:
        """
        True if the host requires interactive actions upon login.
        False by default. Subclasses should override the method to customize the value.
        """
        return False

    @property
    def sanitize_regex(self) -> re.Pattern:
        """
        Regular expression to sanitize sensitive info in command outputs.
        """
        if not self._sanitize_regex:
            escaped_key = re.escape(SANITIZE_KEY)
            # Optionally match the newline that comes from the "echo" command.
            # The -n option for echo to suppress the newline seems to not be
            # supported on all systems
            self._sanitize_regex = re.compile(
                f"{escaped_key}\r?\n?(.*?)(?:{escaped_key}\r?\n?|$)", re.DOTALL
            )

        return self._sanitize_regex

    def sanitize_command(self, cmd: str) -> str:
        """
        Sanitizes a command by adding a prefix and suffix to the command string if
        sanitization is enabled.
        The prefix and suffix are the same and are used to mark the parts of the output
        that should be sanitized. The prefix and suffix are defined by `SANITIZE_KEY`.

        Parameters
        ----------
        cmd
            The command string to be sanitized

        Returns
        -------
        str
            The sanitized command string
        """
        if self.sanitize:
            echo_cmd = f'echo "{SANITIZE_KEY}" | tee /dev/stderr'
            cmd = f"{echo_cmd};{cmd};{echo_cmd}"
        return cmd

    def sanitize_output(self, output: str) -> str:
        """
        Sanitizes the output of a command by selecting the section between the
        SANITIZE_KEY strings.
        If the second instance of the key is not found, the part of the output after the key is returned.
        If the key is not present, the entire output is returned.

        Parameters
        ----------
        output
            The output of the command to be sanitized

        Returns
        -------
        str
            The sanitized output
        """
        if self.sanitize:
            match = self.sanitize_regex.search(output)
            if not match:
                logger.warning(
                    f"Even if sanitization was required, there was no match for the output: {output}. Returning the complete output"
                )
                return output
            return match.group(1)
        return output


class HostError(Exception):
    pass



from typing import Annotated, Optional

import typer

from jobflow_remote.cli.jf import app


@app.command()
def gui(
    port: Annotated[
        Optional[int],
        typer.Option(
            "--port",
            "-p",
            help="The port where the web server will be started. Defaults to 5001",
        ),
    ] = None,
) -> None:
    """
    Start the server for the GUI
    """
    from jobflow_remote.webgui.webgui import start_gui

    start_gui(port=port)



from __future__ import annotations

from collections import Counter
from datetime import datetime, timezone
from math import ceil
from pathlib import Path
from zoneinfo import ZoneInfo

from monty.dev import requires

try:
    import fasthtml
    from fasthtml.common import (
        H1,
        H3,
        H4,
        A,
        Button,
        Card,
        CheckboxX,
        Dialog,
        Div,
        Favicon,
        Form,
        Group,
        Img,
        Input,
        Label,
        Li,
        Link,
        Main,
        Option,
        P,
        Script,
        Select,
        Span,
        Table,
        Td,
        Th,
        Title,
        Tr,
        Ul,
        fast_app,
        serve,
    )
except ImportError:
    fasthtml = None

    # fake rt decorator
    def rt_fake(*args, **kwargs):
        def wrapper(func):
            return func

        return wrapper

    def fast_app(*args, **kwargs):
        return lambda x: x, rt_fake

    def fake_function(*args, **kwargs):
        return None

    Title = Link = Script = Favicon = fake_function


from jobflow_remote import ConfigManager
from jobflow_remote.jobs.daemon import DaemonManager, DaemonStatus
from jobflow_remote.jobs.graph import get_mermaid
from jobflow_remote.jobs.jobcontroller import JobController
from jobflow_remote.jobs.report import JobsReport
from jobflow_remote.jobs.state import JobState

id_curr = "current-info"
id_list = "info-list"


PAGE_TITLE = Title("Jobflow remote manager")

mermaid_js = """
import mermaid from 'https://cdn.jsdelivr.net/npm/mermaid@11/dist/mermaid.esm.min.mjs';
mermaid.initialize({ startOnLoad: false });
window.mermaid = mermaid;
"""


def error_handler(req, exc):
    error_message = f"{exc}"
    return Dialog(
        Div(
            H3("An error occurred:"),
            P(f"{error_message}"),
            Button(
                "Cancel",
                hx_get="/test/close_dialog",
                hx_target="#dialog-container",
                style="font-weight: bold",
            ),
            cls="card-dialog",
        ),
        id="my-dialog",
        open="open",  # This attribute opens the dialog
        cls="dialog",
    )


exception_handlers = {500: error_handler}

app, rt = fast_app(
    pico=False,  # Disable Pico CSS to use custom styles
    hdrs=(
        Link(rel="stylesheet", href="./style.css", type="text/css"),
        Script(mermaid_js, type="module"),
        Favicon("jfr_favicon.ico", "jfr_favicon.ico"),
    ),  # Add custom CSS as a header
    exception_handlers=exception_handlers,
    static_path=Path(__file__).parent,
)


# javascript
js_timezone = """
document.getElementById('timezone_in').value = Intl.DateTimeFormat().resolvedOptions().timeZone;
"""

# change color on click
js_set_color = """
  function setActiveLink(clickedLink) {
    // Remove active class from all links
    document.querySelectorAll('.navbar-link').forEach(link => {
      link.classList.remove('active');
    });

    // Add active class to clicked link
    clickedLink.classList.add('active');
  }
"""

start_stop_btn_lbl = {
    "RUNNING": "Stop",
    "SHUT_DOWN": "Start",
    "STOPPED": "Start",
}

status_colors = {
    DaemonStatus.STOPPED: "red",
    DaemonStatus.STOPPING: "aqua",
    DaemonStatus.SHUT_DOWN: "red",
    DaemonStatus.PARTIALLY_RUNNING: "lawngreen",
    DaemonStatus.STARTING: "aqua",
    DaemonStatus.RUNNING: "limegreen",
}

cm = ConfigManager()

list_projects = list(cm.projects.keys())

job_controllers = {}
daemon_managers = {}
job_controller = None
job_controller_actions = None
daemon_manager = None


def set_job_controller_deamon(project_name):
    global job_controller, job_controller_actions, daemon_manager  # # noqa: PLW0603

    if project_name not in job_controllers:
        job_controllers[project_name] = JobController.from_project_name(
            project_name=project_name
        )
        daemon_managers[project_name] = DaemonManager.from_project(
            cm.get_project(project_name)
        )

    daemon_manager = daemon_managers[project_name]
    job_controller = job_controllers[project_name]
    job_controller_actions = {
        "jobs": {
            "Play": job_controller.play_jobs,
            "Pause": job_controller.pause_jobs,
            "Stop": job_controller.stop_jobs,
            "Retry": job_controller.retry_jobs,
            "Rerun": job_controller.rerun_jobs,
        },
        "flows": {
            "Delete": job_controller.delete_flows,
        },
    }


jfreport = None


def update_jfreport(interval: str = "days", ni: int = 7):
    global jfreport  # noqa: PLW0603

    jfreport = JobsReport().generate_report(
        job_controller, interval=interval, num_intervals=ni
    )


# Navigation bar component
def projectbar(proj_name: str = "", what: str = ""):
    if proj_name:
        status, color = get_runner_status(proj_name)
    return Div(
        Script(js_set_color),
        Ul(
            A(Img(src="./logo_jfr.png", height=50), href="/"),
            Li("Projects:"),
            Li(
                Select(
                    Option("Pick one"),
                    *[
                        Option(prj, value=prj, selected=(proj_name == prj))
                        for prj in list_projects
                    ],
                    name="proj_name",
                    hx_push_url="true",
                    hx_get="/projects",
                    hx_target="body",
                )
            ),
            (
                Li(
                    A(
                        "Report",
                        hx_get=f"/projects?proj_name={proj_name}",
                        hx_push_url="true",
                        hx_target="body",
                        # _="on click remove .active from .navbar a then add .active to me",
                        onclick="setActiveLink(this)",
                        cls="navbar-link",
                    )
                ),
                Group(
                    Ul(
                        Li("Query: "),
                        Li(
                            A(
                                "Jobs",
                                hx_get=f"/{proj_name}/jobs/query",
                                hx_target="#prj-container",
                                hx_push_url="true",
                                # _="on click remove .active from .navbar a then add .active to me",
                                onclick="setActiveLink(this)",
                                cls="navbar-link",
                            )
                        ),
                        Li(
                            A(
                                "Flows",
                                hx_get=f"/{proj_name}/flows/query",
                                hx_target="#prj-container",
                                hx_push_url="true",
                                # _="on click remove .active from .navbar a then add .active to me",
                                onclick="setActiveLink(this)",
                                cls="navbar-link",
                            )
                        ),
                    ),
                    cls="group",
                ),
                Group(
                    Ul(
                        Li("Runner:"),
                        Li(
                            f"{status}",
                            style=f"color: {color}; background-color: black; padding: 3px 5px; border-radius: 4px;",
                        ),
                        Li(
                            Button(
                                f"{start_stop_btn_lbl[status]}",
                                hx_post=f"/runner/{proj_name}/{start_stop_btn_lbl[status].lower()}",
                                hx_target="#runner-status",
                            )
                        ),
                    ),
                    hx_get=f"/runner/{proj_name}/status",
                    hx_trigger="every 30s",
                    hx_swap="outerHTML",
                    id="runner-status",
                    cls="group",
                ),
            )
            if proj_name
            else None,
        ),
        cls="navbar",
    )


def ScrollableArea(*content, height="300px"):
    return Div(
        *content,
        style=f"height: {height}; overflow-y: auto; border: 1px solid #ccc; padding: 10px;",
    )


@rt("/")
def get_home():
    return PAGE_TITLE, Main(
        projectbar(),
        Div(
            H4(
                "Select project from the navigation bar above to view and query Jobs and Flows."
            ),
            P("Add here some general information, e.g. the list of available projects"),
            cls="container",
            id="prj-container",
        ),
        id="page",
    )


@rt("/actions/{action}/{what}/open_dialog", methods=["POST"])
def open_dialog(action: str, what: str, kwargs: dict):
    selected = [f"{v}" for k, v in kwargs.items()]

    # Add checkboxes for delete_flow options
    delete_options = ""
    if action == "Delete" and what == "flows":
        delete_options = Div(
            CheckboxX(
                id="delete_output",
                name="delete_options",
                value="delete_output",
                label="Delete Output",
            ),
            CheckboxX(
                id="delete_files",
                name="delete_options",
                value="delete_files",
                label="Delete Files",
            ),
        )

    return Dialog(
        Div(
            H3(f"Action Selected: {action}"),
            P(f"{what} selected:"),
            P(", ".join(selected) if selected else "No options selected"),
            P("Options:"),
            delete_options,
            Button(
                "Cancel",
                hx_get="/test/close_dialog",
                hx_target="#dialog-container",
                style="",
            ),
            Button(
                "Confirm",
                hx_post=f"/actions/{action}/{what}/run",
                hx_include="[name='ckbx_action'],[name='delete_options']",
                hx_target="#dialog-container",
                hx_swap="innerHTML",
                style="font-weight: bold",
            )
            if selected
            else None,
            cls="card-dialog",
        ),
        id="my-dialog",
        open="open",  # This attribute opens the dialog
        cls="dialog",
    )


@rt("/actions/{action}/{what}/run", methods=["POST"])
def run_action(action: str, what: str, kwargs: dict):
    selected = [v for k, v in kwargs.items() if k != "delete_options"]
    delete_options = kwargs.get("delete_options", [])

    if action == "Delete" and what == "flows":
        delete_output = "delete_output" in delete_options
        delete_files = "delete_files" in delete_options
        response = job_controller_actions[what][action](
            selected, delete_output=delete_output, delete_files=delete_files
        )
    else:
        response = job_controller_actions[what][action](db_ids=selected)

    return Dialog(
        Div(
            H3(f"Action Selected: {action}"),
            P(f"Applied on {response} {what}"),
            P(f"{delete_options}"),
            Button(
                "Cancel",
                hx_get="/test/close_dialog",
                hx_target="#dialog-container",
                style="font-weight: bold",
            ),
            cls="card-dialog",
        ),
        id="my-dialog",
        open="open",  # This attribute opens the dialog
        cls="dialog",
    )


@rt("/test/close_dialog")
def close_dialog():
    return ""


# handle starting and stopping the runner
@rt("/runner/{proj_name}/start", methods=["POST"])
def start_runner_route(proj_name: str):
    dm = daemon_managers[proj_name]
    dm.start()
    status = "STARTING"
    color = status_colors[DaemonStatus(status)]
    return Group(
        Ul(
            Li("Runner:"),
            Li(
                status,
                style=f"color: {color}; background-color: black; padding-right: 10px; border-radius: 4px;",
            ),
        ),
        id="runner-status",
        hx_get=f"/runner/{proj_name}/status",
        hx_trigger="every 5s",
        hx_swap="outerHTML",
        cls="group",
    )


@rt("/runner/{proj_name}/stop", methods=["POST"])
def stop_runner_route(proj_name: str):
    dm = daemon_managers[proj_name]
    dm.shut_down()
    status = "STOPPING"
    color = status_colors[DaemonStatus(status)]
    return Group(
        Ul(
            Li("Runner:"),
            Li(
                status,
                style=f"color: {color}; background-color: black; padding-right: 10px; border-radius: 4px;",
            ),
        ),
        id="runner-status",
        hx_get=f"/runner/{proj_name}/status",
        hx_trigger="every 5s",
        hx_swap="outerHTML",
        cls="group",
    )


@rt("/runner/{proj_name}/status")
def get_runner_status_update(proj_name: str):
    status, color = get_runner_status(proj_name)
    return Group(
        Ul(
            Li("Runner:"),
            Li(
                f"{status}",
                style=f"color: {color}; background-color: black; padding-right: 10px; border-radius: 4px;",
            ),
            Li(
                Button(
                    f"{start_stop_btn_lbl[status]}",
                    hx_post=f"/runner/{proj_name}/{start_stop_btn_lbl[status].lower()}",
                    hx_target="#runner-status",
                )
            ),
        ),
        hx_get=f"/runner/{proj_name}/status",
        hx_trigger="every 30s",
        hx_swap="outerHTML",
        id="runner-status",
        cls="group",
    )


def get_runner_status(proj_name: str):
    dm = daemon_managers[proj_name]
    current_status = dm.check_status()
    color = status_colors[current_status]

    return current_status.name, color


@rt("/projects")
def get_proj_home(proj_name: str = ""):
    if proj_name not in list_projects:
        return PAGE_TITLE, Main(
            projectbar(),
            Div(
                H1(f"The project: {proj_name} does not exists"),
                P("Select an available project from the navigation bar above."),
                id="prj-container",
            ),
        )

    set_job_controller_deamon(proj_name)

    if jfreport:
        interval = jfreport.trends.interval
        ni = jfreport.trends.num_intervals
    else:
        interval = "days"
        ni = 7

    return PAGE_TITLE, Main(
        Script(mermaid_js, type="module"),
        projectbar(proj_name),
        Div(
            H3(f"Report for the Project: {proj_name}"),
            P("Description of the project"),
            Div(
                H3("Jobs Report"),
                Ul(
                    Li(
                        Button(
                            "Summary Jobs Report",
                            hx_get=f"/{proj_name}/jobs/sum_report",
                            hx_target="#jobs-report",
                        )
                    ),
                    Li(
                        Button(
                            "Jobs State Distribution",
                            hx_get=f"/{proj_name}/jobs/state_distro",
                            hx_target="#jobs-report",
                        )
                    ),
                    Li(
                        Form(
                            Button(
                                "Jobs Trend for the last",
                                hx_post=f"/{proj_name}/jobs/trends/",
                                hx_include="[name='interval'],[name='ni']",
                                hx_target="#jobs-report",
                            ),
                            Select(
                                *[
                                    Option(i, selected=i == interval)
                                    for i in ("days", "hours", "months", "years")
                                ],
                                name="interval",
                            ),
                            Input(type="number", value=ni, name="ni"),
                        )
                    ),
                ),
                Div(id="jobs-report"),
                id="buttons-jobs-report",
                cls="card",
            ),
            Div(
                H3("Flows Report"),
                Ul(
                    Li(
                        Button(
                            "Summary Flows Report",
                            hx_get=f"/{proj_name}/flows/sum_report",
                            hx_target="#flows-report",
                        )
                    ),
                    Li(
                        Button(
                            "Flows State Distribution",
                            hx_get=f"/{proj_name}/flows/state_distro",
                            hx_target="#flows-report",
                        )
                    ),
                    Li(
                        Form(
                            Button(
                                "Flows Trend for the last",
                                hx_post=f"/{proj_name}/flows/trends/",
                                hx_include="[name='interval'],[name='ni']",
                                hx_target="#flows-report",
                            ),
                            Select(
                                *[
                                    Option(i, selected=i == interval)
                                    for i in ("days", "hours", "months", "years")
                                ],
                                name="interval",
                            ),
                            Input(type="number", value=ni, name="ni"),
                        )
                    ),
                ),
                Div(id="flows-report"),
                id="buttons-flows-report",
                cls="card",
            ),
            id="prj-container",
        ),
    )


@rt("/{proj_name}/{what}/sum_report")
def sum_report(proj_name: str, what: str):
    if not jfreport:
        update_jfreport()

    state_counts = Counter(jfreport.state_counts)

    # Find the most common state
    most_common_state, most_common_count = state_counts.most_common(1)[0]

    running_count = jfreport.running
    completed_count = jfreport.completed
    error_count = jfreport.error
    active_count = jfreport.active

    # Calculate percentages
    total_jobs = state_counts.total()  # type: ignore
    state_percentages = {
        state: (count / total_jobs) * 100 for state, count in state_counts.items()
    }

    # Create the report
    report = Div(
        Button(
            "Update",
            hx_get=f"/{proj_name}/{what}/sum_report",
            hx_target=f"#{what}-report",
        ),
        Table(
            Tr(Td(f"Total number of {what}:"), Td(f"{total_jobs}")),
            Tr(
                Td("Most common state:"),
                Td(
                    f"{most_common_state.name} ({most_common_count} {what}, {state_percentages[most_common_state]:.2f}%)"
                ),
            ),
            Tr(Td(f"Running {what}:"), Td(f"{running_count}")),
            Tr(Td(f"Completed {what}:"), Td(f"{completed_count}")),
            Tr(
                Td(f"Sum of failed, remote error, and paused {what}:"),
                Td(f"{error_count}"),
            ),
            Tr(Td(f"Sum of all active {what}:"), Td(f"{active_count}")),
            Tr(Td("Longest running:"), Td(f"{jfreport.longest_running}")),
            Tr(Td("Worker utilization:"), Td(f"{jfreport.worker_utilization}")),
        ),
        cls="job-state-report",
    )

    return Div(report, cls="container")


@rt("/{proj_name}/{what}/trends/", methods=["POST"])
def trends(proj_name: str, what: str, interval: str = "days", ni: int = 7):
    if not jfreport:
        update_jfreport()

    if jfreport.trends.interval != interval or jfreport.trends.num_intervals != ni:
        update_jfreport(interval, ni)

    # trends table
    trend_table = Table(
        Tr(Th("Dates"), Th("Completed"), Th("Failed"), Th("Remote error")),
        *[
            Tr(Td(d), Td(c), Td(f), Td(r))
            for d, c, f, r in zip(
                jfreport.trends.dates,
                jfreport.trends.completed,
                jfreport.trends.failed,
                jfreport.trends.remote_error,
            )
        ],
        cls="trend-table",
    )

    report = Div(
        H3(f"{what.capitalize()} Trend Table:"), trend_table, cls="job-state-report"
    )

    return Div(report, cls="container")


@rt("/{proj_name}/{what}/state_distro")
def state_distro(proj_name: str, what: str):
    if not jfreport:
        update_jfreport()

    state_counts = Counter(jfreport.state_counts)

    # Calculate percentages
    total_jobs = state_counts.total()  # type: ignore
    state_percentages = {
        state: (count / total_jobs) * 100 for state, count in state_counts.items()
    }

    # Create the report
    report = Div(
        Table(
            Tr(Th("State"), Th("Count"), Th("Percentage")),
            *[
                Tr(
                    Td(state.name),
                    Td(str(count)),
                    Td(f"{state_percentages[state]:.2f}%"),
                )
                for state, count in state_counts.items()
            ],
            cls="state-distribution-table",
        ),
        cls="job-state-report",
    )
    return Div(
        Button(
            "Update",
            hx_get=f"/{proj_name}/{what}/state_distro",
            hx_target=f"#{what}-report",
        ),
        report,
        cls="container",
    )


@rt("/{proj_name}/{what}/info/{jf_id}")
def get_info_job_flow(jf_id: str, what: str, proj_name: str):
    info = None

    if what == "jobs":
        info = job_controller.get_job_info(db_id=jf_id).dict()
    elif what == "flows":
        info = job_controller.get_flow_info_by_flow_uuid(jf_id)

    if info:
        return Div(
            H3(f"Job {jf_id} details"),
            H3(f"Job name: {info.pop('name')}"),
            ScrollableArea(
                *[
                    Ul(
                        Li(f"{k}:", style="font-weight: bold"),
                        Li(f"{v}", style="margin-left: 10px; al"),
                    )
                    for k, v in info.items()
                ]
            ),
            cls="card-dialog active",
        )

    return P("Job not found")


@rt("/{proj_name}/flows/graph/{jf_id}")
def get_graph_job_flow(jf_id: str, proj_name: str):
    flowinfo = job_controller.get_flows_info(limit=1, full=True)[0]
    graph = get_mermaid(flowinfo)
    m_script = f"""
(async function() {{
    const container = document.getElementById('flow-graph');
    const {{ svg }} = await window.mermaid.render('graphDiv', `{graph}`);
    container.innerHTML = svg;
}})();
"""
    return Div(
        ScrollableArea(Div(id="flow-graph"), Script(m_script)), cls="card-dialog"
    )


@rt("/{proj_name}/{what}/dialog/{jf_id}")
def get_info_graph_dialog(jf_id: str, what: str, proj_name: str):
    return Dialog(
        Card(
            Button(
                "Close",
                hx_get="/test/close_dialog",
                hx_target="#dialog-container",
                style="float: right; font-weight: bold",
                cls="btn",
            ),
            Div(
                Button(
                    f"Details {what}",
                    cls="tab active",
                    hx_get=f"/{proj_name}/{what}/info/{jf_id}",
                    hx_target="#tab-content",
                    hx_swap="innerHTML",
                    _="on click remove .active from .tab then add .active to me",
                ),
                Button(
                    "Graph Flow",
                    cls="tab",
                    hx_get=f"/{proj_name}/flows/graph/{jf_id}",
                    hx_target="#tab-content",
                    hx_swap="innerHTML",
                    _="on click remove .active from .tab then add .active to me",
                ),
                cls="tab-buttons btn",
            )
            if what == "flows"
            else None,
            Div(
                get_info_job_flow(jf_id, what, proj_name),
                id="tab-content",
                cls="tab-content",
            ),
            cls="tabbed-card",
        ),
        id="my-dialog",
        open="open",
        cls="dialog",
    )


@rt("/{proj_name}/{what}/close_info")
def close_info():
    return ""


@rt("/{proj_name}/{what}/query")
def get(proj_name: str, what: str):
    form = Form(
        Group(
            Input(name="db_id", placeholder="DB ID")
            if what == "jobs"
            else Input(name="flow_id", placeholder="Flow ID"),
            Input(name="uuid", placeholder="UUID") if what == "jobs" else None,
            Input(name="name", placeholder="Job Name"),
            cls="group",
        ),
        Group(
            Label("State"),
            Select(
                Option("Any", value=""),
                *[Option(state.name, value=state.value) for state in JobState],
                name="state",
                id="select-job-state",
            ),
            Input(name="worker", placeholder="Worker") if what == "jobs" else None,
            cls="group",
        ),
        # Date and time for start and end time
        Group(
            Label("Start Date/Time"),
            Input(type="date", name="start_date"),
            Input(type="time", name="start_time", label="Start Time"),
            Label("End Date/Time"),
            Input(type="date", name="end_date", label="End Date"),
            Input(type="time", name="end_time", label="End Time"),
            cls="group",
        ),
        # Input for entries per page
        Group(
            Label("Entries per page"),
            Select(
                *[
                    Option(str(i), value=str(i), selected=(i == 20))
                    for i in (10, 20, 50, 100)
                ],
                name="entries_per_page",
                id="select-npages",
            ),
            cls="group",
        ),
        # Hidden input to store the user's timezone (populated by JavaScript)
        Input(type="hidden", name="timezone_in", value="", id="timezone_in"),
        Input(type="hidden", name="sort_by", value="updated_on"),
        Input(type="hidden", name="sort_order", value="-1"),
        Button("Search", cls="btn"),
        Script(js_timezone),
        Script(mermaid_js, type="module"),
        hx_post=f"/{proj_name}/{what}/query",
        hx_target="#query-results",
        cls="card",
        id="search-form",
    )

    if what == "jobs":
        total_entries = job_controller.count_jobs()
    elif what == "flows":
        total_entries = job_controller.count_flows()

    return Div(
        H3(f"{what.capitalize()} Query"),
        H3(f"Total number of {what}: {total_entries}"),
        form,
        Div(id="query-results"),
        id="prj-container",
    )


@rt("/{proj_name}/{what}/query")
def post(
    proj_name: str,
    what: str,
    db_id: str = "",
    uuid: str = "",
    name: str = "",
    state: str = "",
    worker: str = "",
    start_date: str = "",
    start_time: str = "",
    end_date: str = "",
    end_time: str = "",
    max_results: int = 100,
    timezone_in: str = "",
    sort_by: str = "updated_on",
    sort_order: int = -1,
    page: int = 1,
    entries_per_page: int = 10,
):
    query: dict[
        str,
        str
        | list[str]
        | tuple[str, None]
        | list[tuple[str, None]]
        | None
        | JobState
        | dict[str, str]
        | datetime,
    ] = {}
    if db_id:
        query["db_ids"] = [db_id]
    if uuid:
        query["job_ids"] = [(uuid, None)]
    if name:
        query["name"] = name
    if state:
        query["states"] = None if state == "Any" else JobState(state)
    if worker:
        query["metadata"] = {"worker": worker}

    if timezone_in:
        tz = ZoneInfo(timezone_in)
    else:
        # TODO this should be fixed
        # tz = ZoneInfo(datetime.now(timezone.utc).astimezone().tzname())
        raise NotImplementedError

    if start_time and not start_date:
        start_date = datetime.now().strftime("%Y-%m-%d")
    if start_date:
        if start_time:
            start_date = f"{start_date} {start_time}"
            query["start_date"] = datetime.strptime(start_date, "%Y-%m-%d %H:%M")
        else:
            query["start_date"] = datetime.strptime(start_date, "%Y-%m-%d")

    if end_time and not end_date:
        end_date = datetime.now().strftime("%Y-%m-%d")
    if end_date:
        if end_time:
            end_date = f"{end_date} {end_time}"
            query["end_date"] = datetime.strptime(end_date, "%Y-%m-%d %H:%M")
        else:
            query["end_date"] = datetime.strptime(end_date, "%Y-%m-%d")

    skip = (page - 1) * entries_per_page

    if what == "jobs":
        jobs_flows = job_controller.get_jobs_info(
            **query, sort=[[sort_by, sort_order]], limit=entries_per_page, skip=skip
        )
    elif what == "flows":
        jobs_flows = job_controller.get_flows_info(
            **query, sort=[[sort_by, sort_order]], limit=entries_per_page, skip=skip
        )

    if not jobs_flows:
        return Div(P("No jobs found matching the criteria."), cls="card")

    # Toggle sort direction for next click
    next_sort_order = -sort_order

    # prepare all entries
    all_entries = list(jobs_flows)

    # prepare table
    info_table = Table(
        Tr(
            Th(
                A(
                    "DB id",
                    hx_post=f"/{proj_name}/{what}/query",
                    hx_include="previous form",
                    hx_vals=f'{{"sort_by":"db_id", "sort_order":"{next_sort_order if sort_by == "db_id" else 1}"}}',
                    hx_swap="innerHTML",
                    hx_target="#query-results",
                )
            )
            if what == "jobs"
            else None,
            Th(
                A(
                    "UUID",
                    hx_post=f"/{proj_name}/{what}/query",
                    hx_include="previous form",
                    hx_vals=f'{{"sort_by":"uuid", "sort_order":"{next_sort_order if sort_by == "uuid" else 1}"}}',
                    hx_swap="innerHTML",
                    hx_target="#query-results",
                )
            ),
            Th(
                A(
                    "Name",
                    hx_post=f"/{proj_name}/{what}/query",
                    hx_include="previous form",
                    hx_vals=f'{{"sort_by":"job.name", "sort_order":"{next_sort_order if sort_by == "job.name" else 1}"}}',
                    hx_swap="innerHTML",
                    hx_target="#query-results",
                )
            ),
            Th(
                A(
                    "State",
                    hx_post=f"/{proj_name}/{what}/query",
                    hx_include="previous form",
                    hx_vals=f'{{"sort_by":"state", "sort_order":"{next_sort_order if sort_by == "state" else 1}"}}',
                    hx_swap="innerHTML",
                    hx_target="#query-results",
                )
            ),
            Th(
                A(
                    "Worker",
                    hx_post=f"/{proj_name}/{what}/query",
                    hx_include="previous form",
                    hx_vals=f'{{"sort_by":"worker", "sort_order":"{next_sort_order if sort_by == "worker" else 1}"}}',
                    hx_swap="innerHTML",
                    hx_target="#query-results",
                )
            )
            if what == "jobs"
            else None,
            Th(
                A(
                    "Updated",
                    hx_post=f"/{proj_name}/{what}/query",
                    hx_include="previous form",
                    hx_vals=f'{{"sort_by":"updated_on", "sort_order":"{next_sort_order if sort_by == "updated_on" else 1}"}}',
                    hx_swap="innerHTML",
                    hx_target="#query-results",
                )
            ),
            Th(
                Label("Action", Input(type="checkbox", onclick="toggleAll(this)")),
                style="text-align: right",
            ),
        ),
        *[
            Tr(
                Td(
                    A(
                        entry.db_id,
                        hx_get=f"/{proj_name}/{what}/dialog/{entry.db_id}",
                        hx_target="#dialog-container",
                        hx_swap="innerHTML",
                    )
                )
                if what == "jobs"
                else None,
                Td(entry.uuid)
                if what == "jobs"
                else Td(
                    A(
                        entry.flow_id,
                        hx_get=f"/{proj_name}/{what}/dialog/{entry.flow_id}",
                        hx_target="#dialog-container",
                        hx_swap="innerHTML",
                    )
                ),
                Td(entry.name),
                Td(entry.state.value),
                Td(entry.worker) if what == "jobs" else None,
                Td(
                    entry.updated_on.replace(tzinfo=timezone.utc)
                    .astimezone(tz)
                    .strftime("%Y-%m-%d %H:%M:%S")
                ),
                Td(
                    Input(
                        type="checkbox",
                        name="ckbx_action",
                        value=f"{entry.db_id if what=='jobs' else entry.flow_id}",
                        id=f"{entry.db_id if what=='jobs' else entry.flow_id}",
                    ),
                    style="text-align: right",
                ),
                style="background-color: #f5f5f5;" if i % 2 == 0 else "",
            )
            for i, entry in enumerate(all_entries)
        ],
        cls="card",
    )

    if what == "jobs":
        total_entries = job_controller.count_jobs(**query)
    elif what == "flows":
        total_entries = job_controller.count_flows(**query)

    total_pages = ceil(total_entries / entries_per_page)

    if what == "jobs":
        # all_job_flow_ids = [entry.db_id for entry in all_entries]
        action_btns_lbl = ("Rerun", "Play", "Pause", "Stop", "Retry")
    elif what == "flows":
        # all_job_flow_ids = [entry.flow_id for entry in all_entries]
        action_btns_lbl = ("Delete",)  # type: ignore[assignment]

    action_btns = Group(
        Label("Actions:   "),
        *[
            Button(
                name,
                hx_post=f"/actions/{name}/{what}/open_dialog",
                hx_target="#dialog-container",
                hx_include="[name='ckbx_action']",
                cls="btn",
            )
            for name in action_btns_lbl
        ],
        style="float: right",
        cls="group",
    )

    pagination = Div(
        A(
            "Previous",
            hx_post=f"/{proj_name}/{what}/query",
            hx_include="previous form",
            hx_vals=f'{{"page":{page-1}}}',
            hx_swap="innerHTML",
            hx_target="#query-results",
        )
        if page > 1
        else Span("Previous"),
        *[
            A(
                str(i),
                hx_post=f"/{proj_name}/{what}/query",
                hx_include="previous form",
                hx_vals=f'{{"page":{i}}}',
                hx_swap="innerHTML",
                hx_target="#query-results",
            )
            for i in range(max(1, page - 2), min(total_pages + 1, page + 3))
        ],
        A(
            "Next",
            hx_post=f"/{proj_name}/{what}/query",
            hx_include="previous form",
            hx_vals=f'{{"page":{page+1}}}',
            hx_swap="innerHTML",
            hx_target="#query-results",
        )
        if page < total_pages
        else Span("Next"),
        cls="pagination",
    )

    toggle_all_ckbx = Script("""
        function toggleAll(source) {
        var checkboxes = document.querySelectorAll('input[type="checkbox"][name="ckbx_action"]');
        for (var i = 0; i < checkboxes.length; i++) {checkboxes[i].checked = source.checked;}}
    """)
    return Div(
        H4(f"Total after filter:{total_entries}"),
        action_btns,
        info_table,
        pagination,
        toggle_all_ckbx,
        id="query-results",
        cls="card",
    ), Div(Script(mermaid_js, type="module"), id="dialog-container")


@requires(
    fasthtml is not None, "The 'python-fasthtml' package is required to run the gui."
)
def start_gui(port: int | None = None):
    serve(
        appname="jobflow_remote.webgui.webgui",
        port=port,
        reload_includes=[Path(__file__).parent],
    )


if __name__ == "__main__":
    serve()



from __future__ import annotations

import os
from collections.abc import Mapping, MutableMapping
from copy import deepcopy
from datetime import datetime, timezone
from typing import Any
from uuid import UUID

import maggma.stores  # required to enable subclass searching
from dateutil.relativedelta import relativedelta
from dateutil.tz import gettz
from maggma.core.store import Store
from monty.json import MontyDecoder


def deep_merge_dict(
    d1: MutableMapping,
    d2: Mapping,
    path: list[str] | None = None,
    raise_on_conflicts: bool = True,
    inplace: bool = True,
) -> MutableMapping:
    """
    Merge a dictionary d2 into a dictionary d1 recursively.

    Parameters
    ----------
    d1
    d2
    path
    raise_on_conflicts
    inplace

    Returns
    -------

    """
    if not inplace:
        d1 = deepcopy(d1)
    if path is None:
        path = []
    for key in d2:
        if key in d1:
            if isinstance(d1[key], Mapping) and isinstance(d2[key], Mapping):
                deep_merge_dict(d1[key], d2[key], [*path, str(key)])
            elif d1[key] == d2[key]:
                pass  # same leaf value
            elif raise_on_conflicts:
                raise ValueError(f"Conflict at {'.'.join([*path, str(key)])}")
            else:
                d1[key] = d2[key]
        else:
            d1[key] = d2[key]
    return d1


def remove_none(obj):
    if isinstance(obj, (list, tuple, set)):
        return type(obj)(remove_none(x) for x in obj if x is not None)
    if isinstance(obj, dict):
        return type(obj)(
            (remove_none(k), remove_none(v))
            for k, v in obj.items()
            if k is not None and v is not None
        )
    return obj


def check_dict_keywords(obj: Any, keywords: list[str]) -> bool:
    if isinstance(obj, (list, tuple, set)):
        return any(check_dict_keywords(x, keywords) for x in obj)
    if isinstance(obj, dict):
        for k, v in obj.items():
            if isinstance(k, str) and any(k.startswith(kw) for kw in keywords):
                return True
            if check_dict_keywords(v, keywords):
                return True
    return False


def uuid_to_path(
    uuid: str, index: int | None = 1, num_subdirs: int = 3, subdir_len: int = 2
):
    u = UUID(uuid)
    u_hex = u.hex

    # Split the digest into groups of "subdir_len" characters
    subdirs = [
        u_hex[i : i + subdir_len]
        for i in range(0, num_subdirs * subdir_len, subdir_len)
    ]

    # add the index to the final dir name
    dir_name = f"{uuid}"
    if index is not None:
        dir_name += f"_{index}"

    # Combine root directory and subdirectories to form the final path
    return os.path.join(*subdirs, dir_name)


def store_from_dict(store_dict: dict) -> Store:
    if "@class" in store_dict and "@module" in store_dict:
        store = MontyDecoder().process_decoded(store_dict)
        if not isinstance(store, Store):
            raise ValueError(
                f"The converted object {store} is not an instance of a maggma Store"
            )
        return store

    def all_subclasses(cl):
        return set(cl.__subclasses__()).union(
            [s for c in cl.__subclasses__() for s in all_subclasses(c)]
        )

    all_stores = {s.__name__: s for s in all_subclasses(maggma.stores.Store)}
    return convert_store(store_dict, all_stores)


def convert_store(spec_dict: dict, valid_stores) -> Store:
    """
    Build a store based on the dict spec configuration from JobFlow
    TODO expose the methods from jobflow and don't duplicate the code.
    """
    _spec_dict = dict(spec_dict)
    store_type = _spec_dict.pop("type")
    for k, v in _spec_dict.items():
        if isinstance(v, dict) and "type" in v:
            _spec_dict[k] = convert_store(v, valid_stores)
    return valid_stores[store_type](**_spec_dict)


def convert_utc_time(datetime_value: datetime) -> datetime:
    """
    Convert a time in UTC (used in the DB) to the time zone of the
    system where the code is being executed.

    Parameters
    ----------
    datetime_value
        a datetime object in UTC

    Returns
    -------
        The datetime in the zone of the current system
    """
    return datetime_value.replace(tzinfo=timezone.utc).astimezone(tz=None)


def get_past_time_rounded(
    interval: str, num_intervals: int, reference: datetime | None = None
) -> datetime:
    """
    Return a datetime object that is the specified number of intervals in the
    past relative to the given reference datetime. The returned datetime is
    rounded to the nearest interval start time.

    Parameters
    ----------
    interval
        One of 'hours', 'days', 'weeks', 'months', 'years'
    num_intervals
        The number of intervals to go back in time
    reference
        The datetime to use as the reference for the calculation. If not
        specified, the current time is used.

    Returns
    -------
    datetime
        The datetime object that is the specified number of intervals in the
        past relative to the given reference datetime.
    """
    if not reference:
        reference = datetime.utcnow()
    past = reference - relativedelta(**{interval: num_intervals - 1})  # type: ignore[arg-type]

    # Define starting point modifications based on quantity type
    start_modifiers = {
        "hours": lambda dt: dt.replace(minute=0, second=0, microsecond=0),
        "days": lambda dt: dt.replace(hour=0, minute=0, second=0, microsecond=0),
        "weeks": lambda dt: dt.replace(hour=0, minute=0, second=0, microsecond=0)
        - relativedelta(days=dt.weekday()),
        "months": lambda dt: dt.replace(
            day=1, hour=0, minute=0, second=0, microsecond=0
        ),
        "years": lambda dt: dt.replace(
            month=1, day=1, hour=0, minute=0, second=0, microsecond=0
        ),
    }

    return start_modifiers[interval](past)


def get_utc_offset(timezone: str):
    # Get current time in the specified timezone
    """
    Return the UTC offset of the given timezone as a string.

    Parameters
    ----------
    timezone
        The timezone for which to get the UTC offset.

    Returns
    -------
    str
        The UTC offset as a string in the format +/-HH:MM.
    """
    tz_info_val = gettz(timezone)
    if not tz_info_val:
        raise ValueError(f"Could not determine the timezone for {timezone}")
    now = datetime.now(tz_info_val)

    # Get the UTC offset
    utc_offset = now.utcoffset()

    # Extract hours and minutes
    hours, remainder = divmod(utc_offset.total_seconds(), 3600)
    minutes = remainder // 60

    return f"{int(hours):+03d}:{int(minutes):02d}"


# TODO imported this from jobflow remote for backward compatibility.
# remove this in favor of suid in the next release
def suuid() -> str:
    """
    Generate a string UUID (universally unique identifier).

    Uses the UUID4 specification.

    Returns
    -------
    str
        A UUID.
    """
    from uuid import uuid4

    return str(uuid4())



from __future__ import annotations

from typing import TYPE_CHECKING, Callable

from jobflow_remote.config import ConfigManager

if TYPE_CHECKING:
    from jobflow import Flow, Job, JobStore
    from qtoolkit.core.data_objects import QResources

    from jobflow_remote.config.base import ExecutionConfig


def set_run_config(
    flow_or_job: Flow | Job,
    name_filter: str = None,
    function_filter: Callable = None,
    exec_config: str | ExecutionConfig | None = None,
    resources: dict | QResources | None = None,
    priority: int | None = None,
    worker: str | None = None,
    dynamic: bool = True,
) -> Flow | Job:
    """
    Modify in place a Flow or a Job by setting the properties in the
    "manager_config" entry in the JobConfig associated to each Job
    matching the filter.

    The values left as None will not be set and their value may be
    filled by those set in the submit_flow.
    If you want/need to leave resources or exec_config empty pass
    an empty QResources (or an empty dict) or an empty ExecutionConfig.

    Uses the Flow/Job update_config() method,
    so follows the same conventions, also setting the options in
    the config_updates of the Job, to allow setting the same properties
    also in dynamically generated Jobs, if the dynamic option is True.

    Note that calling this function will override all previously set
    values for matching Jobs, even if not specified here.

    Parameters
    ----------
    flow_or_job
        A Flow or a Job to be modified
    name_filter
        A filter for the job name. Only jobs with a matching name will be updated.
        Includes partial matches, e.g. "ad" will match a job with the name "adder".
    function_filter
        A filter for the job function. Only jobs with a matching function will be
        updated.
    exec_config
        The execution configuration to be added to the selected Jobs.
    resources
        The resources to be set for the selected Jobs.
    priority
        The priority of the Job.
    worker
        The worker where the selected Jobs will be executed.
    dynamic
        The updates will be propagated to Jobs/Flows dynamically generated at
        runtime.

    Returns
    -------
    Flow or Job
        The modified object.
    """
    if not exec_config and not resources and not worker:
        return flow_or_job
    config: dict = {"manager_config": {}}
    if exec_config is not None:
        config["manager_config"]["exec_config"] = exec_config
    if resources is not None:
        config["manager_config"]["resources"] = resources
    if worker is not None:
        config["manager_config"]["worker"] = worker
    if priority is not None:
        config["manager_config"]["priority"] = priority

    flow_or_job.update_config(
        config=config,
        name_filter=name_filter,
        function_filter=function_filter,
        dynamic=dynamic,
    )

    return flow_or_job


def load_job_store(project: str | None = None) -> JobStore:
    """
    Load the JobStore for the current project.

    Parameters
    ----------
    project

    Returns
    -------

    """
    cm = ConfigManager()
    p = cm.get_project(project)
    return p.get_jobstore()



from __future__ import annotations

import glob
import logging
import os
import shutil
import traceback
from pathlib import Path
from typing import TYPE_CHECKING, NamedTuple

import tomlkit
from monty.json import jsanitize
from monty.os import makedirs_p
from monty.serialization import dumpfn, loadfn

from jobflow_remote.config.base import (
    ConfigError,
    ExecutionConfig,
    Project,
    ProjectUndefinedError,
    WorkerBase,
)
from jobflow_remote.utils.data import deep_merge_dict

if TYPE_CHECKING:
    from jobflow import JobStore
    from maggma.stores import MongoStore

logger = logging.getLogger(__name__)


class ProjectData(NamedTuple):
    filepath: str | Path
    project: Project
    ext: str


class WorkerData(NamedTuple):
    name: str
    worker: str


class ConfigManager:
    """
    A manager for the projects configuration files.

    Provides tool to parse project information from the selected projects folder as
    well as methods to update the properties of each project.
    """

    projects_ext = ("json", "yaml", "toml")

    def __init__(
        self,
        exclude_unset: bool = False,
        exclude_none: bool = False,
        warn: bool = False,
    ) -> None:
        """

        Parameters
        ----------
        exclude_unset
            when dumping projects determine whether fields which were not explicitly
            set when creating the model should be excluded from the dictionary
        exclude_none
            when dumping projects determine whether fields which are equal to None
            should be excluded from the dictionary
        warn
            if True print warnings related to the parsing of the files in the
            projects folder
        """
        from jobflow_remote import SETTINGS

        self.exclude_unset = exclude_unset
        self.exclude_none = exclude_none
        self.warn = warn
        self.projects_folder = Path(SETTINGS.projects_folder)
        self.projects_data = self.load_projects_data()

    @property
    def projects(self) -> dict[str, Project]:
        """
        Returns
        -------
        dict
            Dictionary with project name as key and Project as value.
        """
        return {name: pd.project for name, pd in self.projects_data.items()}

    def load_projects_data(self) -> dict[str, ProjectData]:
        """
        Load projects from the selected projects folder.

        Returns
        -------
        dict
            Dictionary with project name as key and ProjectData as value.
        """
        projects_data: dict[str, ProjectData] = {}
        for ext in self.projects_ext:
            for filepath in self.projects_folder.glob(str(f"*.{ext}")):
                try:
                    if ext in ["json", "yaml"]:
                        d = loadfn(filepath)
                    else:
                        with open(filepath) as f:
                            d = tomlkit.parse(f.read())
                    project = Project.parse_obj(d)
                except Exception:
                    if self.warn:
                        logger.warning(
                            f"File {filepath} could not be parsed as a Project. Error: {traceback.format_exc()}"
                        )
                    continue
                if project.name in projects_data:
                    msg = f"Two projects with the same name '{project.name}' have been defined: {filepath}, {projects_data[project.name].filepath}"
                    raise ConfigError(msg)
                projects_data[project.name] = ProjectData(filepath, project, ext)

        return projects_data

    def select_project_name(self, project_name: str | None = None) -> str:
        """
        Determine the project name to be used based on the passed value
        and on the general settings.

        Parameters
        ----------
        project_name
            The name of the project or None to use the value from the settings

        Returns
        -------
        str
            The name of the selected project.
        """
        from jobflow_remote import SETTINGS

        project_name = project_name or SETTINGS.project
        if not project_name:
            if len(self.projects_data) == 1:
                project_name = next(iter(self.projects_data))
            else:
                raise ProjectUndefinedError(
                    f"A project name should be defined, known projects: {list(self.projects_data)}"
                )

        return project_name

    def get_project_data(self, project_name: str | None = None) -> ProjectData:
        """
        Get the ProjectData object based from the project name.

        Parameters
        ----------
        project_name
            The name of the project or None to use the value from the settings

        Returns
        -------
        ProjectData
            The selected ProjectData
        """
        project_name = self.select_project_name(project_name)

        if project_name not in self.projects_data:
            raise ConfigError(
                f"The selected project {project_name} does not exist "
                "or could not be parsed correctly"
            )

        return self.projects_data[project_name]

    def get_project(self, project_name: str | None = None) -> Project:
        """
        Get the Project object based from the project name.

        Parameters
        ----------
        project_name
            The name of the project or None to use the value from the settings

        Returns
        -------
        Project
            The selected Project
        """
        return self.get_project_data(project_name).project

    def dump_project(self, project_data: ProjectData) -> None:
        """
        Dump the project to filepath specified in the ProjectData.

        Parameters
        ----------
        project_data
            The project data to be dumped
        """
        exclude_none = True if project_data.ext == "toml" else self.exclude_none
        d = jsanitize(
            project_data.project.dict(
                exclude_none=exclude_none, exclude_unset=self.exclude_unset
            ),
            enum_values=True,
        )
        if project_data.ext in ["json", "yaml"]:
            dumpfn(d, project_data.filepath)
        elif project_data.ext == "toml":
            with open(project_data.filepath, "w") as f:
                tomlkit.dump(d, f)

    def create_project(self, project: Project, ext="yaml") -> None:
        """
        Create a new Project in the project folder by dumping the project to file.

        Parameters
        ----------
        project
            The data of the project to be created.
        ext
            The extension of the file to which the project will be dumped (yaml, json
             or toml)
        """
        if project.name in self.projects_data:
            raise ConfigError(f"Project with name {project.name} already exists")

        makedirs_p(project.base_dir)
        makedirs_p(project.tmp_dir)
        makedirs_p(project.log_dir)
        filepath = self.projects_folder / f"{project.name}.{ext}"
        if filepath.exists():
            raise ConfigError(
                f"Project with name {project.name} does not exist, but file {filepath!s} does"
            )
        project_data = ProjectData(filepath, project, ext)
        self.dump_project(project_data)
        self.projects_data[project.name] = project_data

    def remove_project(self, project_name: str, remove_folders: bool = True) -> None:
        """
        Remove a project from the projects folder.

        Parameters
        ----------
        project_name
            Name of the project to be removed.
        remove_folders
            Optionally remove the folders related to the project (e.g. tmp, log).
        """
        if project_name not in self.projects_data:
            return
        project_data = self.projects_data.pop(project_name)
        if remove_folders:
            shutil.rmtree(project_data.project.base_dir, ignore_errors=True)
        os.remove(project_data.filepath)

    def update_project(self, config: dict, project_name: str) -> None:
        """
        Update the project values.
        The passed dict with values will be recursively merged in the current project.

        Parameters
        ----------
        config
            Dictionary with the project values to be updated.
        project_name
            Name of the project to be updated
        """
        project_data = self.projects_data.pop(project_name)
        proj_dict = project_data.project.dict()
        new_project = Project.parse_obj(deep_merge_dict(proj_dict, config))
        project_data = ProjectData(project_data.filepath, new_project, project_data.ext)
        self.dump_project(project_data)
        self.projects_data[project_data.project.name] = project_data

    def project_names_from_files(self) -> list[str]:
        """
        Parses all the prasable files and only checks for the "name" attribute to
        return a list of potential project file names.

        Useful in case some projects cannot be properly parsed, but the full list
        needs to be returned.

        Returns
        -------
        list
            List of project names.
        """
        project_names = []
        for ext in self.projects_ext:
            for filepath in glob.glob(str(self.projects_folder / f"*.{ext}")):
                try:
                    if ext in ["json", "yaml"]:
                        d = loadfn(filepath)
                    else:
                        with open(filepath) as f:
                            d = tomlkit.parse(f.read())
                    if "name" in d:
                        project_names.append(d["name"])
                except Exception:
                    logger.warning(
                        f"File {filepath} could not be parsed as a Project. Error: {traceback.format_exc()}"
                    )
                    continue

        return project_names

    def set_worker(
        self,
        name: str,
        worker: WorkerBase,
        project_name: str | None = None,
        replace: bool = False,
    ) -> None:
        """
        Set a worker in the selected project.
        Can add a new worker or replace an existing one.

        Parameters
        ----------
        name
            Name of the worker to be added or replaced.
        worker
            Worker to be set.
        project_name
            Name of the project where the Worker is set, or None to use the one
            from the settings.
        replace
            Raise an exception if False and a Worker with the chosen name already
            exists.
        """
        project_data = self.get_project_data(project_name)
        if not replace and name in project_data.project.workers:
            raise ConfigError(f"Worker with name {name} is already defined")

        project_data.project.workers[name] = worker
        self.dump_project(project_data)

    def remove_worker(self, worker_name: str, project_name: str | None = None) -> None:
        """
        Remove a worker from the selected project.

        Parameters
        ----------
        worker_name
            Name of the worker to be removed
        project_name
            Name of the project from which the Worker should be removed, or None to
            use the one from the settings.
        """
        project_data = self.get_project_data(project_name)
        project_data.project.workers.pop(worker_name)
        self.dump_project(project_data)

    def get_worker(
        self, worker_name: str, project_name: str | None = None
    ) -> WorkerBase:
        """
        Return the worker object based on the name.

        Parameters
        ----------
        worker_name
            Name of the worker to retrieve.
        project_name
            Name of the project from which the Worker should be retrieved, or None to
            use the one from the settings.

        Returns
        -------
        WorkerBase
            The selected Worker.
        """
        project = self.get_project(project_name)
        if worker_name not in project.workers:
            raise ConfigError(f"Worker with name {worker_name} is not defined")
        return project.workers[worker_name]

    def set_queue_db(self, store: MongoStore, project_name: str | None = None) -> None:
        """
        Set the project specific store used for managing the queue.

        Parameters
        ----------
        store
            A maggma Store
        project_name
            Name of the project where the Store is set, or None to use the one
            from the settings.
        """
        project_data = self.get_project_data(project_name)
        project_data.project.queue = store.as_dict()

        self.dump_project(project_data)

    def set_jobstore(self, jobstore: JobStore, project_name: str | None = None) -> None:
        """
        Set the project specific store used for jobflow.

        Parameters
        ----------
        jobstore
            A maggma Store
        project_name
            Name of the project where the Store is set, or None to use the one
            from the settings.
        """
        project_data = self.get_project_data(project_name)
        project_data.project.jobstore = jobstore.as_dict()
        self.dump_project(project_data)

    def set_exec_config(
        self,
        exec_config_name: str,
        exec_config: ExecutionConfig,
        project_name: str | None = None,
        replace: bool = False,
    ) -> None:
        """
        Set an ExecutionConfig in the selected project.
        Can add a new ExecutionConfig or replace an existing one.

        Parameters
        ----------
        exec_config_name
            Name of the ExecutionConfig to be added or replaced.
        exec_config
            The ExecutionConfig.
        project_name
            Name of the project where the ExecutionConfig is set, or None to use
             the one from the settings.
        replace
            Raise an exception if False and an ExecutionConfig with the chosen
            name already exists.
        """
        project_data = self.get_project_data(project_name)
        if not replace and exec_config_name in project_data.project.exec_config:
            raise ConfigError(f"Host with name {exec_config_name} is already defined")
        project_data.project.exec_config[exec_config_name] = exec_config
        self.dump_project(project_data)

    def remove_exec_config(
        self, exec_config_name: str, project_name: str | None = None
    ) -> None:
        """
        Remove an ExecutionConfig from the selected project.

        Parameters
        ----------
        exec_config_name
            Name of the ExecutionConfig to be removed
        project_name
            Name of the project from which the ExecutionConfig should be removed, or
            None to use the one from the settings.
        """
        project_data = self.get_project_data(project_name)
        project_data.project.exec_config.pop(exec_config_name, None)
        self.dump_project(project_data)

    def get_exec_config(
        self, exec_config_name: str, project_name: str | None = None
    ) -> ExecutionConfig:
        """
        Return the ExecutionConfig object based on the name.

        Parameters
        ----------
        exec_config_name
            Name of the ExecutionConfig.
        project_name
            Name of the project from which the ExecutionConfig should be retrieved,
            or None to use the one from the settings.

        Returns
        -------
        ExecutionConfig
            The selected ExecutionConfig
        """
        project = self.get_project(project_name)
        if exec_config_name not in project.exec_config:
            raise ConfigError(
                f"ExecutionConfig with id {exec_config_name} is not defined"
            )
        return project.exec_config[exec_config_name]
